# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
HAL component providing access to and decoding of ACPI tables
"""

__version__ = '0.1'

import struct
from collections import defaultdict, namedtuple
from collections.abc import Callable
from typing import Dict, List, Tuple, Optional, Union, TYPE_CHECKING

from chipsec.library.defines import bytestostring
from chipsec.library.exceptions import UnimplementedAPIError
from chipsec.library.file import read_file
from chipsec.hal import acpi_tables
from chipsec.hal.hal_base import HALBase
from chipsec.hal.uefi import UEFI
from chipsec.library.logger import logger, print_buffer_bytes
from chipsec.hal.acpi_tables import ACPI_TABLE

if TYPE_CHECKING:
    from ctypes import Array

# ACPI Table Header Format
ACPI_TABLE_HEADER_FORMAT = '=4sIBB6s8sI4sI'
ACPI_TABLE_HEADER_SIZE = struct.calcsize(ACPI_TABLE_HEADER_FORMAT)  # 36
assert 36 == ACPI_TABLE_HEADER_SIZE


class ACPI_TABLE_HEADER(namedtuple('ACPI_TABLE_HEADER', 'Signature Length Revision Checksum OEMID OEMTableID OEMRevision CreatorID CreatorRevision')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""  Table Header
------------------------------------------------------------------
  Signature        : {self.Signature}
  Length           : 0x{self.Length:08X}
  Revision         : 0x{self.Revision:02X}
  Checksum         : 0x{self.Checksum:02X}
  OEM ID           : {self.OEMID}
  OEM Table ID     : {self.OEMTableID}
  OEM Revision     : 0x{self.OEMRevision:08X}
  Creator ID       : {self.CreatorID}
  Creator Revision : 0x{self.CreatorRevision:08X}
"""


ACPI_TABLE_SIG_SIZE = 0x4

ACPI_TABLE_SIG_ROOT = 'ROOT'
ACPI_TABLE_SIG_RSDP = 'RSDP'
ACPI_TABLE_SIG_RSDT = 'RSDT'
ACPI_TABLE_SIG_XSDT = 'XSDT'
ACPI_TABLE_SIG_FACP = 'FACP'
ACPI_TABLE_SIG_FACS = 'FACS'
ACPI_TABLE_SIG_DSDT = 'DSDT'
ACPI_TABLE_SIG_SSDT = 'SSDT'
ACPI_TABLE_SIG_PSDT = 'PSDT'
ACPI_TABLE_SIG_APIC = 'APIC'
ACPI_TABLE_SIG_SBST = 'SBST'
ACPI_TABLE_SIG_ECDT = 'ECDT'
ACPI_TABLE_SIG_SRAT = 'SRAT'
ACPI_TABLE_SIG_SLIC = 'SLIC'
ACPI_TABLE_SIG_SLIT = 'SLIT'
ACPI_TABLE_SIG_BOOT = 'BOOT'
ACPI_TABLE_SIG_CPEP = 'CPEP'
ACPI_TABLE_SIG_DBGP = 'DBGP'
ACPI_TABLE_SIG_ETDT = 'ETDT'
ACPI_TABLE_SIG_HPET = 'HPET'
ACPI_TABLE_SIG_MCFG = 'MCFG'
ACPI_TABLE_SIG_SPCR = 'SPCR'
ACPI_TABLE_SIG_SPMI = 'SPMI'
ACPI_TABLE_SIG_TCPA = 'TCPA'
ACPI_TABLE_SIG_WDAT = 'WDAT'
ACPI_TABLE_SIG_WDRT = 'WDRT'
ACPI_TABLE_SIG_WSPT = 'WSPT'
ACPI_TABLE_SIG_WDDT = 'WDDT'
ACPI_TABLE_SIG_ASF = 'ASF!'
ACPI_TABLE_SIG_MSEG = 'MSEG'
ACPI_TABLE_SIG_DMAR = 'DMAR'
ACPI_TABLE_SIG_UEFI = 'UEFI'
ACPI_TABLE_SIG_FPDT = 'FPDT'
ACPI_TABLE_SIG_PCCT = 'PCCT'
ACPI_TABLE_SIG_MSDM = 'MSDM'
ACPI_TABLE_SIG_BATB = 'BATB'
ACPI_TABLE_SIG_BGRT = 'BGRT'
ACPI_TABLE_SIG_LPIT = 'LPIT'
ACPI_TABLE_SIG_ASPT = 'ASPT'
ACPI_TABLE_SIG_FIDT = 'FIDT'
ACPI_TABLE_SIG_HEST = 'HEST'
ACPI_TABLE_SIG_BERT = 'BERT'
ACPI_TABLE_SIG_ERST = 'ERST'
ACPI_TABLE_SIG_EINJ = 'EINJ'
ACPI_TABLE_SIG_TPM2 = 'TPM2'
ACPI_TABLE_SIG_WSMT = 'WSMT'
ACPI_TABLE_SIG_DBG2 = 'DBG2'
ACPI_TABLE_SIG_NHLT = 'NHLT'
ACPI_TABLE_SIG_MSCT = 'MSCT'
ACPI_TABLE_SIG_RASF = 'RASF'
ACPI_TABLE_SIG_OEM1 = 'OEM1'
ACPI_TABLE_SIG_OEM2 = 'OEM2'
ACPI_TABLE_SIG_OEM3 = 'OEM3'
ACPI_TABLE_SIG_OEM4 = 'OEM4'
ACPI_TABLE_SIG_NFIT = 'NFIT'

ACPI_TABLES: Dict[str, Callable] = {
    ACPI_TABLE_SIG_ROOT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_RSDT: acpi_tables.RSDT,
    ACPI_TABLE_SIG_XSDT: acpi_tables.XSDT,
    ACPI_TABLE_SIG_FACP: acpi_tables.FADT,
    ACPI_TABLE_SIG_FACS: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_DSDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_SSDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_PSDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_APIC: acpi_tables.APIC,
    ACPI_TABLE_SIG_SBST: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_ECDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_SRAT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_SLIC: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_SLIT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_BOOT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_CPEP: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_DBGP: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_ETDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_HPET: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_MCFG: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_SPCR: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_TCPA: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_WDAT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_WDRT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_WSPT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_WDDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_ASF: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_MSEG: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_DMAR: acpi_tables.DMAR,
    ACPI_TABLE_SIG_UEFI: acpi_tables.UEFI_TABLE,
    ACPI_TABLE_SIG_FPDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_PCCT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_MSDM: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_BATB: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_BGRT: acpi_tables.BGRT,
    ACPI_TABLE_SIG_LPIT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_ASPT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_FIDT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_HEST: acpi_tables.HEST,
    ACPI_TABLE_SIG_BERT: acpi_tables.BERT,
    ACPI_TABLE_SIG_ERST: acpi_tables.ERST,
    ACPI_TABLE_SIG_EINJ: acpi_tables.EINJ,
    ACPI_TABLE_SIG_TPM2: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_WSMT: acpi_tables.WSMT,
    ACPI_TABLE_SIG_DBG2: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_NHLT: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_MSCT: acpi_tables.MSCT,
    ACPI_TABLE_SIG_RASF: acpi_tables.RASF,
    ACPI_TABLE_SIG_SPMI: acpi_tables.SPMI,
    ACPI_TABLE_SIG_OEM1: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_OEM2: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_OEM3: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_OEM4: acpi_tables.ACPI_TABLE,
    ACPI_TABLE_SIG_NFIT: acpi_tables.NFIT
}

########################################################################################################
#
# RSDP
#
########################################################################################################

RSDP_GUID_ACPI2_0 = '8868E871-E4F1-11D3-BC22-0080C73C8881'
RSDP_GUID_ACPI1_0 = 'EB9D2D31-2D88-11D3-9A16-0090273FC14D'
ACPI_RSDP_SIG = 'RSD PTR '

########################################################################################################
#
# ACPI HAL Component
#
########################################################################################################


class ACPI(HALBase):
    def __init__(self, cs):
        super(ACPI, self).__init__(cs)
        self.uefi = UEFI(self.cs)
        self.tableList: Dict[str, List[int]] = defaultdict(list)
        self.get_ACPI_table_list()

    def read_RSDP(self, rsdp_pa: int) -> acpi_tables.RSDP:
        rsdp_buf = self.cs.mem.read_physical_mem(rsdp_pa, acpi_tables.ACPI_RSDP_SIZE)
        rsdp = acpi_tables.RSDP()
        rsdp.parse(rsdp_buf)
        if rsdp.Revision >= 0x2:
            rsdp_buf = self.cs.mem.read_physical_mem(rsdp_pa, acpi_tables.ACPI_RSDP_EXT_SIZE)
            rsdp = acpi_tables.RSDP()
            rsdp.parse(rsdp_buf)
        return rsdp

    #
    # Check RSDP in Extended BIOS Data Area
    #
    def _find_RSDP_in_EBDA(self) -> Tuple[Optional[acpi_tables.RSDP], Optional[int]]:
        rsdp_pa = None
        rsdp = None
        logger().log_hal('[acpi] searching RSDP in EBDA...')
        ebda_ptr_addr = 0x40E
        ebda_addr = struct.unpack('<H', self.cs.mem.read_physical_mem(ebda_ptr_addr, 2))[0] << 4
        if ebda_addr > 0x400 and ebda_addr < 0xA0000:
            membuf = self.cs.mem.read_physical_mem(ebda_addr, 0xA0000 - ebda_addr)
            pos = bytestostring(membuf).find(ACPI_RSDP_SIG)
            if -1 != pos:
                rsdp_pa = ebda_addr + pos
                rsdp = self.read_RSDP(rsdp_pa)
                if rsdp.is_RSDP_valid():
                    logger().log_hal(f'[acpi] found RSDP in EBDA at: 0x{rsdp_pa:016X}')
                else:
                    rsdp_pa = None
        return rsdp, rsdp_pa

    #
    # Search RSDP in legacy BIOS E/F segments (0xE0000 - 0xFFFFF)
    #
    def _find_RSDP_in_legacy_BIOS_segments(self) -> Tuple[Optional[acpi_tables.RSDP], Optional[int]]:
        rsdp_pa = None
        rsdp = None
        membuf = self.cs.mem.read_physical_mem(0xE0000, 0x20000)
        membuf = bytestostring(membuf)
        pos = bytestostring(membuf).find(ACPI_RSDP_SIG)
        if -1 != pos:
            rsdp_pa = 0xE0000 + pos
            rsdp = self.read_RSDP(rsdp_pa)
            if rsdp.is_RSDP_valid():
                logger().log_hal(f'[acpi] Found RSDP in BIOS E/F segments: 0x{rsdp_pa:016X}')
            else:
                rsdp_pa = None
        return rsdp, rsdp_pa

    #
    # Search for RSDP in the EFI memory (EFI Configuration Table)
    #
    def _find_RSDP_in_EFI_config_table(self) -> Tuple[Optional[acpi_tables.RSDP], Optional[int]]:
        rsdp_pa = None
        rsdp = None
        logger().log_hal('[acpi] Searching RSDP pointers in EFI Configuration Table...')
        (isFound, _, ect, _) = self.uefi.find_EFI_Configuration_Table()
        if isFound and (ect is not None):
            if RSDP_GUID_ACPI2_0 in ect.VendorTables:
                rsdp_pa = ect.VendorTables[RSDP_GUID_ACPI2_0]
                logger().log_hal(f'[acpi] ACPI 2.0+ RSDP {{{RSDP_GUID_ACPI2_0}}} in EFI Config Table: 0x{rsdp_pa:016X}')
            elif RSDP_GUID_ACPI1_0 in ect.VendorTables:
                rsdp_pa = ect.VendorTables[RSDP_GUID_ACPI1_0]
                logger().log_hal('[acpi] ACPI 1.0 RSDP {{{RSDP_GUID_ACPI1_0}}} in EFI Config Table: 0x{rsdp_pa:016X}')

        if rsdp_pa:
            rsdp = self.read_RSDP(rsdp_pa)
            if rsdp.is_RSDP_valid():
                logger().log_hal(f'[acpi] Found RSDP in EFI Config Table: 0x{rsdp_pa:016X}')
            else:
                rsdp_pa = None
        return rsdp, rsdp_pa

    #
    # Search for RSDP in all EFI memory
    #
    def _find_RSDP_in_EFI(self) -> Tuple[Optional[acpi_tables.RSDP], Optional[int]]:
        rsdp_pa = None
        rsdp = None
        logger().log_hal("[acpi] Searching all EFI memory for RSDP (this may take a minute).")
        CHUNK_SZ = 1024 * 1024  # 1MB
        (smram_base, _, _) = self.cs.cpu.get_SMRAM()
        pa = smram_base - CHUNK_SZ
        while pa > CHUNK_SZ:
            membuf = self.cs.mem.read_physical_mem(pa, CHUNK_SZ)
            pos = bytestostring(membuf).find(ACPI_RSDP_SIG)
            if -1 != pos:
                rsdp_pa = pa + pos
                logger().log_hal(f"[acpi] Found '{ACPI_RSDP_SIG}' signature at 0x{rsdp_pa:16X}. Checking if valid RSDP.")
                rsdp = self.read_RSDP(rsdp_pa)
                if rsdp.is_RSDP_valid():
                    logger().log_hal(f'[acpi] Found RSDP in EFI memory: 0x{rsdp_pa:016X}')
                    break
            pa -= CHUNK_SZ
        return rsdp, rsdp_pa

    #
    # Searches for Root System Description Pointer (RSDP) in various locations for legacy/EFI systems
    #
    def find_RSDP(self) -> Tuple[Optional[int], Optional[acpi_tables.RSDP]]:
        rsdp, rsdp_pa = self._find_RSDP_in_EBDA()

        if rsdp_pa is None:
            rsdp, rsdp_pa = self._find_RSDP_in_legacy_BIOS_segments()

        if rsdp_pa is None:
            rsdp, rsdp_pa = self._find_RSDP_in_EFI_config_table()

        if rsdp_pa is None:
            rsdp, rsdp_pa = self._find_RSDP_in_EFI()

        if rsdp is not None:
            logger().log_hal(str(rsdp))

        return (rsdp_pa, rsdp)

    RsdtXsdt = Union[acpi_tables.RSDT, acpi_tables.XSDT]

    #
    # Retrieves System Description Table (RSDT or XSDT) either from RSDP or using OS API
    #
    def get_SDT(self, search_rsdp: bool = True) -> Tuple[bool, Optional[int], Optional[RsdtXsdt], Optional[ACPI_TABLE_HEADER]]:
        is_xsdt = False
        sdt_pa = None
        sdt_header = None
        sdt_buf = b''
        if search_rsdp:
            (_, rsdp) = self.find_RSDP()
            if rsdp is not None:
                if 0x0 == rsdp.Revision:
                    sdt_pa = rsdp.RsdtAddress
                    is_xsdt = False
                elif 0x2 == rsdp.Revision:
                    sdt_pa = rsdp.XsdtAddress
                    is_xsdt = True
                else:
                    return (False, None, None, None)
                found_str = 'XSDT' if is_xsdt else 'RSDT'
                logger().log_hal(f'[acpi] Found {found_str} at PA: 0x{sdt_pa:016X}')
                sdt_header_buf = self.cs.mem.read_physical_mem(sdt_pa, ACPI_TABLE_HEADER_SIZE)
                sdt_header = self._parse_table_header(sdt_header_buf)
                sdt_buf = self.cs.mem.read_physical_mem(sdt_pa, sdt_header.Length)
        else:
            sdt_pa = None
            if logger().HAL:
                logger().log("[acpi] Reading RSDT/XSDT using OS API...")
            (sdt_buf, is_xsdt) = self.get_ACPI_SDT()
            sdt_header = self._parse_table_header(sdt_buf[:ACPI_TABLE_HEADER_SIZE])

        sdt_contents = sdt_buf[ACPI_TABLE_HEADER_SIZE:]
        sdt = ACPI_TABLES[ACPI_TABLE_SIG_XSDT if is_xsdt else ACPI_TABLE_SIG_RSDT]()
        sdt.parse(sdt_contents)
        return (is_xsdt, sdt_pa, sdt, sdt_header)

    def get_ACPI_SDT(self) -> Tuple[Optional['Array'], bool]:
        sdt = self.cs.helper.get_acpi_table('XSDT')  # FirmwareTableID_XSDT
        xsdt = sdt is not None
        if not xsdt:
            sdt = self.get_acpi_table('RSDT')  # FirmwareTableID_RSDT
        return sdt, xsdt

    #
    # Populates a list of ACPI tables available on the system
    #
    def get_ACPI_table_list(self) -> Dict[str, List[int]]:
        try:
            # 1. If didn't work, try using get_ACPI_table if a helper implemented
            #    reading ACPI tables via native API which some OS may provide
            # raise UnimplementedAPIError("asdf")
            logger().log_hal("[acpi] Trying to enumerate ACPI tables using get_ACPI_table...")
            for table_name in self.cs.helper.enum_ACPI_tables():
                self.tableList[table_name.decode("utf-8")].append(0)
        except UnimplementedAPIError:
            # 2. Try to extract ACPI table(s) from physical memory
            #    read_physical_mem can be implemented using both
            #    CHIPSEC kernel module and OS native API
            logger().log_hal("[acpi] Trying to enumerate ACPI tables from physical memory...")
            # find RSDT/XSDT table
            (is_xsdt, sdt_pa, sdt, sdt_header) = self.get_SDT()

            # cache RSDT/XSDT in the list of ACPI tables
            if (sdt_pa is not None) and (sdt_header is not None):
                self.tableList[bytestostring(sdt_header.Signature)].append(sdt_pa)
            if sdt is not None:
                self.get_table_list_from_SDT(sdt, is_xsdt)
            self.get_DSDT_from_FADT()

        return self.tableList

    #
    # Gets table list from entries in RSDT/XSDT
    #
    def get_table_list_from_SDT(self, sdt: RsdtXsdt, is_xsdt: bool) -> None:
        logger().log_hal(f'[acpi] Getting table list from entries in {"XSDT" if is_xsdt else "RSDT"}')
        for a in sdt.Entries:
            _sig = self.cs.mem.read_physical_mem(a, ACPI_TABLE_SIG_SIZE)
            _sig = bytestostring(_sig)
            if _sig not in ACPI_TABLES.keys():
                if logger().HAL:
                    logger().log_warning(f'Unknown ACPI table signature: {_sig}')
            self.tableList[_sig].append(a)

    #
    # Gets DSDT from FADT
    #
    def get_DSDT_from_FADT(self) -> None:
        logger().log_hal('[acpi] Getting DSDT from FADT')

        if ACPI_TABLE_SIG_FACP in self.tableList:
            (_, parsed_fadt_content, _, _) = self.get_parse_ACPI_table('FACP')[0]
        else:
            if logger().HAL:
                found_table = 'XSDT' if ACPI_TABLE_SIG_XSDT in self.tableList else 'RSDT'
                logger().log_warning(f'Cannot find FADT in {found_table}')
            return

        dsdt_address_to_use = parsed_fadt_content.get_DSDT_address_to_use()

        if dsdt_address_to_use is None:
            dsdt_address = parsed_fadt_content.dsdt
            x_dsdt_address = parsed_fadt_content.x_dsdt
            if logger().HAL:
                logger().log_error('Unable to determine the correct DSDT address')
            if logger().HAL:
                logger().log_error(f'  DSDT   address = 0x{dsdt_address:08X}')
            if logger().HAL:
                address_str = f'{x_dsdt_address:16X}' if x_dsdt_address is not None else 'Not found'
                logger().log_error(f'  X_DSDT address = 0x{address_str}')
            return

        self.tableList[ACPI_TABLE_SIG_DSDT].append(dsdt_address_to_use)

    #
    # Checks is ACPI table with <name> is available on the system
    #
    def is_ACPI_table_present(self, name: str) -> bool:
        return (name in self.tableList)

    #
    # Prints a list of ACPI tables available on the system
    #
    def print_ACPI_table_list(self) -> None:
        if len(self.tableList) == 0:
            logger().log_error("Couldn't get a list of ACPI tables")
        else:
            logger().log_hal('[acpi] Found the following ACPI tables:')
            for tableName in sorted(self.tableList.keys()):
                table_values_str = ', '.join([f'0x{addr:016X}' for addr in self.tableList[tableName]])
                logger().log(f' - {tableName}: {table_values_str}')

    #
    # Retrieves contents of ACPI table from memory or from file
    #
    def get_parse_ACPI_table(self, name: str, isfile: bool = False) -> List['ParseTable']:
        acpi_tables = self.get_ACPI_table(name, isfile)
        return [self._parse_table(name, table_header_blob, table_blob) for (table_header_blob, table_blob) in acpi_tables if table_header_blob is not None]

    def get_ACPI_table(self, name: str, isfile: bool = False) -> List[Tuple[bytes, bytes]]:
        acpi_tables_data: List[bytes] = []
        if isfile:
            acpi_tables_data.append(read_file(name))
        else:
            try:
                # 1. Try to extract ACPI table(s) using get_ACPI_table if a helper implemented
                #    reading ACPI tables via native API which some OS may provide
                logger().log_hal("[acpi] trying to extract ACPI table using get_ACPI_table...")
                t_data = self.cs.helper.get_ACPI_table(name)
                if t_data is None:
                    raise TypeError
                acpi_tables_data.append(t_data)
            except UnimplementedAPIError:
                # 2. If didn't work, try scrubbing physical memory
                #    read_physical_mem can be implemented using both
                #    CHIPSEC kernel module and OS native API
                logger().log_hal('[acpi] trying to extract ACPI table from physical memory...')
                for table_address in self.tableList[name]:
                    t_size = self.cs.mem.read_physical_mem_dword(table_address + 4)
                    t_data = self.cs.mem.read_physical_mem(table_address, t_size)
                    acpi_tables_data.append(t_data)

        acpi_tables = []
        for data in acpi_tables_data:
            acpi_tables.append((data[: ACPI_TABLE_HEADER_SIZE], data[ACPI_TABLE_HEADER_SIZE:]))

        return acpi_tables

    #
    # Dumps contents of ACPI table
    #
    def dump_ACPI_table(self, name: str, isfile: bool = False) -> None:
        acpi_tables = self.get_parse_ACPI_table(name, isfile)
        for acpi_table in acpi_tables:
            (table_header, table, table_header_blob, table_blob) = acpi_table
            logger().log("==================================================================")
            logger().log(f'ACPI Table: {name}')
            logger().log("==================================================================")
            # print table header
            logger().log(str(table_header))
            print_buffer_bytes(table_header_blob)
            # print table contents
            logger().log('')
            logger().log(str(table))
            print_buffer_bytes(table_blob)
            logger().log('')

    # --------------------------------------------------------------------
    # Internal ACPI table parsing functions
    # --------------------------------------------------------------------

    ParseTable = Tuple[ACPI_TABLE_HEADER, Optional[ACPI_TABLE], bytes, bytes]

    def _parse_table(self, name: str, table_header_blob: bytes, table_blob: bytes) -> ParseTable:
        table_header = self._parse_table_header(table_header_blob)
        table = self._parse_table_contents(name, table_blob, table_header_blob)
        return (table_header, table, table_header_blob, table_blob)

    def _parse_table_header(self, header: bytes) -> ACPI_TABLE_HEADER:
        acpi_table_hdr = ACPI_TABLE_HEADER(*struct.unpack_from(ACPI_TABLE_HEADER_FORMAT, header))
        logger().log_hal(str(acpi_table_hdr))
        return acpi_table_hdr

    def _parse_table_contents(self, signature: str, contents: bytes, header: bytes) -> Optional[ACPI_TABLE]:
        table = None
        if ACPI_TABLES.__contains__(signature):
            logger().log_hal(f'{signature}')
            if 'BERT' in signature:
                BootRegionLen = struct.unpack('<L', contents[0:4])[0]
                BootRegionAddr = struct.unpack('<Q', contents[4:12])[0]
                bootRegion = self.cs.mem.read_physical_mem(BootRegionAddr, BootRegionLen)
                table = (ACPI_TABLES[signature])(bootRegion)
            elif 'NFIT' in signature:
                table = (ACPI_TABLES[signature])(header)
            else:
                table = (ACPI_TABLES[signature])()
            table.parse(contents)
        return table



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2017, Google Inc
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util tpm parse_log <file>
>>> chipsec_util tpm state <locality>
>>> chipsec_util tpm command <commandName> <locality> <command_parameters>

locality: 0 | 1 | 2 | 3 | 4
commands - parameters:
pccrread - pcr number ( 0 - 23 )
nvread - Index, Offset, Size
startup - startup type ( 1 - 3 )
continueselftest
getcap - Capabilities Area, Size of Sub-capabilities, Sub-capabilities
forceclear

Examples:

>>> chipsec_util tpm parse_log binary_bios_measurements
>>> chipsec_util tpm state 0
>>> chipsec_util tpm command pcrread 0 17
>>> chipsec_util tpm command continueselftest 0
"""

from chipsec.command import BaseCommand, toLoad
from chipsec.hal import tpm_eventlog
from chipsec.hal.tpm import TPM
from chipsec.library.exceptions import TpmRuntimeError
from chipsec.testcase import ExitCode
from argparse import ArgumentParser


class TPMCommand(BaseCommand):

    no_driver_cmd = ['parse_log']

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=__doc__)
        subparsers = parser.add_subparsers()
        parser_parse = subparsers.add_parser('parse_log')
        parser_parse.add_argument('file', type=str, help='File name')
        parser_parse.set_defaults(func=self.tpm_parse)

        parser_command = subparsers.add_parser('command')
        parser_command.add_argument('command_name', type=str, help='Command')
        parser_command.add_argument('locality', type=str, choices=['0', '1', '2', '3', '4'], help='Locality')
        parser_command.add_argument('command_parameters', nargs='*', type=int, help='Command Parameters')
        parser_command.set_defaults(func=self.tpm_command)

        parser_state = subparsers.add_parser('state')
        parser_state.add_argument('locality', type=str, choices=['0', '1', '2', '3', '4'], help='Locality')
        parser_state.set_defaults(func=self.tpm_state)
        parser.parse_args(self.argv, namespace=self)

    def tpm_parse(self):
        with open(self.file, 'rb') as log:
            tpm_eventlog.parse(log)

    def tpm_command(self):
        self._tpm.command(self.command_name, self.locality, self.command_parameters)

    def tpm_state(self):
        self._tpm.dump_all(self.locality)

    def set_up(self):
        if self.func != self.tpm_parse:
            try:
                self._tpm = TPM(self.cs)
            except TpmRuntimeError as msg:
                self.logger.log(msg)
                return

    def run(self):
        try:
            self.func()
        except Exception:
            self.ExitCode = ExitCode.ERROR


commands = {'tpm': TPMCommand}



from unittest import TestCase
from chipsec.library.register import Register, RegisterType
from tests.software.cs import DummyCS

class TestRegister(TestCase):
    def setUp(self):
        self.cs = DummyCS()
        self.reg = Register(self.cs)

    def test_is_defined(self):
        self.assertTrue(self.reg.is_defined('TEST_MSR'))
        self.assertFalse(self.reg.is_defined('NON_EXISTENT'))
        self.assertFalse(self.reg.is_defined('TEST_BAD_MMIO'))
        self.assertTrue(self.reg.is_defined('TEST_MMIO'))

    def test_get_def(self):
        d = self.reg.get_def('TEST_MSR')
        self.assertEqual(d['type'], RegisterType.MSR)
        self.assertIn('FIELDS', d)

    def test_get_field_mask(self):
        mask = self.reg.get_field_mask('TEST_MSR', 'FIELD1')
        self.assertEqual(mask, 0xFF)

    def test_get_field(self):
        value = self.reg.get_field('TEST_MSR', 0x1234, 'FIELD1')
        self.assertEqual(value, 0x34)

    def test_set_field(self):
        val = self.reg.set_field('TEST_MSR', 0xFFFF, 'FIELD1', 0x12)
        self.assertEqual(val & 0xFF, 0x12)

    def test_read_msr(self):
        val = self.reg.read('TEST_MSR')
        self.assertIsInstance(val, int)

    def test_read_all_msr(self):
        vals = self.reg.read_all('TEST_MSR')
        self.assertIsInstance(vals, list)

    def test_write_msr(self):
        self.assertTrue(self.reg.write('TEST_MSR', 0x12345678))

    def test_has_field(self):
        self.assertTrue(self.reg.has_field('TEST_MSR', 'FIELD1'))
        self.assertFalse(self.reg.has_field('TEST_MSR', 'NON_EXISTENT_FIELD'))

    def test_has_all_fields(self):
        self.assertTrue(self.reg.has_all_fields('TEST_MSR', ['FIELD1']))
        self.assertFalse(self.reg.has_all_fields('TEST_MSR', ['FIELD1', 'NON_EXISTENT_FIELD']))

    def test_is_msr(self):
        self.assertIn(self.reg.is_msr('TEST_MSR'), [True, False])

    def test_is_pci(self):
        self.assertIn(self.reg.is_pci('TEST_PCI'), [True, False])

    def test_is_all_ffs(self):
        self.assertIn(self.reg.is_all_ffs('TEST_MSR', 0xFFFFFFFF), [True, False])

    def test_is_field_all_ones(self):
        self.assertTrue(self.reg.is_field_all_ones('TEST_MSR', 'FIELD1', 0xFF))

    def test_read_field(self):
        val = self.reg.read_field('TEST_MSR', 'FIELD1')
        self.assertIsInstance(val, int)

    def test_write_field(self):
        self.assertTrue(self.reg.write_field('TEST_MSR', 'FIELD1', 0xAA))

    def test_read_dict(self):
        d = self.reg.read_dict('TEST_MSR')
        self.assertIn('value', d)

    def test_print(self):
        s = self.reg.print('TEST_MSR', 0x12345678)
        self.assertIsInstance(s, str)


# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Hyper-V VMBus generic fuzzer

Usage:
    ``chipsec_main.py -i -m tools.vmm.hv.vmbusfuzz -a fuzz,<parameters>``

    Parameters:

    - ``all``          : Fuzzing all bytes
    - ``hv``           : Fuzzing HyperV message header
    - ``vmbus``        : Fuzzing HyperV message body / VMBUS message
    - ``<pos>,<size>`` : Fuzzing number of bytes at specific position

Examples:
    >>> chipsec_main.py -i -m tools.vmm.hv.vmbusfuzz -a fuzz,all -l log.txt

.. note::
    - The fuzzer is incompatible with native VMBus driver (``vmbus.sys``). To use it, remove ``vmbus.sys``
    - Returns a Warning by default
    - System may be in an unknown state, further evaluation may be needed

.. important::
    - This module is designed to run in a VM environment
    - Behavior on physical HW is undefined

"""

import sys
import traceback
from struct import pack
from random import getrandbits, choice
from chipsec.library.returncode import ModuleResult
from chipsec.modules.tools.vmm.common import session_logger, overwrite, get_int_arg
from chipsec.modules.tools.vmm.hv.vmbus import VMBusDiscovery, HyperV, RingBuffer
from chipsec.module_common import CPU

sys.stdout = session_logger(True, 'vmbusfuzz')

TAGS = [CPU]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'VMM', 'HV', 'VMBUSFUZZ']


class VMBusFuzz(VMBusDiscovery):
    def __init__(self):
        VMBusDiscovery.__init__(self)
        self.training = False
        self.training_msginfo = []
        self.fuzzing = False
        self.fuzzing_rules = {}
        self.current_message = 0

    ##
    # hv_post_msg - Fuzzing a message to be sent
    ##
    def hv_post_msg(self, message):
        if self.training:
            self.training_msginfo.append(len(message))
        if self.fuzzing:
            if self.current_message in self.fuzzing_rules:
                rules = self.fuzzing_rules[self.current_message]
                for position in rules:
                    message = overwrite(message, rules[position], position)
            self.current_message += 1
        # Randomize leftover bytes. It shouldn't affect functionality.
        leftovers = ''.join(chr(getrandbits(8)) for _ in range(256 - len(message)))
        return HyperV.hv_post_msg(self, message + leftovers)

    def vmbus_test1_run(self):
        self.debug = False
        self.vmbus_request_offers()
        child_relid_list = sorted([value['child_relid'] for value in self.offer_channels.values()])

        if not self.fuzzing:
            for relid in child_relid_list:
                self.ringbuffers[relid] = RingBuffer()
                self.ringbuffers[relid].ringbuffer_alloc(4)
                self.ringbuffers[relid].gpadl = self.vmbus_get_next_gpadl()
                self.ringbuffers[relid].debug = False

        for relid in child_relid_list:
            self.vmbus_establish_gpadl(relid, self.ringbuffers[relid].gpadl, self.ringbuffers[relid].pfn)
            self.vmbus_open(relid, self.ringbuffers[relid].gpadl, self.ringbuffers[relid].send_size)

        if not self.fuzzing:
            self.print_supported_versions()
            self.print_offer_channels()
            self.print_created_gpadl()
            self.print_open_channels()
            self.print_events()

        for relid in child_relid_list:
            self.vmbus_close(relid)
            self.vmbus_teardown_gpadl(relid, self.ringbuffers[relid].gpadl)

        self.vmbus_rescind_all_offers()

    def run(self, module_argv):
        self.logger.start_test('Hyper-V VMBus fuzzer')

        if len(module_argv) > 0:
            command = module_argv[0]
        else:
            self.logger.log(self.__doc__.replace('`', ''))
            return

        cmdarg1 = module_argv[1] if len(module_argv) > 1 else 'all'
        cmdarg2 = module_argv[2] if len(module_argv) > 2 else '1'

        cmdarg1 = get_int_arg(cmdarg1) if cmdarg1 not in ['all', 'hv', 'vmbus'] else cmdarg1
        cmdarg2 = max(1, min(8, get_int_arg(cmdarg2)))

        self.debug = False
        self.promt = 'VMBUS'

        try:
            self.vmbus_init()
            self.scan_supported_versions()
            if len(self.supported_versions):
                version = choice(self.supported_versions.keys())
                self.vmbus_clear()
                self.vmbus_connect(version)
            self.training = True
            self.vmbus_test1_run()
            self.training = False

            if command == 'fuzz':
                m = 0
                for n in self.training_msginfo:
                    range_options = {'all': range(n), 'hv': range(0x10), 'vmbus': range(0x10, n)}
                    fuzzing_range = range_options[cmdarg1] if cmdarg1 in range_options else [cmdarg1]
                    fuzzing_range = [x for x in fuzzing_range if x < n]
                    for i in fuzzing_range:
                        randstr = pack('<Q', getrandbits(64))[:cmdarg2]
                        self.fuzzing_rules = {m: {i: randstr}}
                        self.logger.log(f'[VMBUS] Message: {m + 1:d}/{len(self.training_msginfo):d}  Fuzzing {len(randstr):d} byte(s): position {i:d} out of {n:d}')
                        self.vmbus_clear()
                        if len(self.supported_versions):
                            self.vmbus_connect(version)
                        self.current_message = 0
                        self.fuzzing = True
                        self.vmbus_test1_run()
                        self.fuzzing = False
                    m += 1
        except KeyboardInterrupt:
            self.logger.log('***** Control-C *****')
        except Exception:
            traceback.print_exc()
        finally:
            self.vmbus_rescind_all_offers()

        self.logger.log_information('Module completed')
        self.logger.log_warning('System may be in an unknown state, further evaluation may be needed.')
        self.result.setStatusBit(self.result.status.VERIFY)
        self.res = self.result.getReturnCode(ModuleResult.WARNING)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2020, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Simple CPU Module Specific Register (MSR) VMM emulation fuzzer

Usage:
    ``chipsec_main -m tools.vmm.msr_fuzz [-a random]``

    - ``-a`` : random = use random values (default = sequential numbering)

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -i -m tools.vmm.msr_fuzz
    >>> chipsec_main.py -i -m tools.vmm.msr_fuzz -a random

Additional options set within the module:
    - ``_NO_ITERATIONS_TO_FUZZ`` : Number of iterations to fuzz randomly
    - ``_READ_MSR``              : Specify to read MSR when fuzzing it
    - ``_FLUSH_LOG_EACH_MSR``    : Flush log file before each MSR
    - ``_FUZZ_VALUE_0_all1s``    : Try all 0 & all 1 values to be written to each MSR
    - ``_FUZZ_VALUE_5A``         : Try 0x5A values to be written to each MSR
    - ``_FUZZ_VALUE_RND``        : Try random values to be written to each MSR
    - ``_EXCLUDE_MSR``           : MSR values to exclude (list)

.. note::
    - Returns a Warning by default
    - System may be in an unknown state, further evaluation may be needed

.. important::
    - This module is designed to run in a VM environment
    - Behavior on physical HW is undefined

"""

import random

from chipsec.module_common import BaseModule
from chipsec.library.returncode import ModuleResult

_MODULE_NAME = 'msr_fuzz'

# Number of iterations to fuzz randomly
_NO_ITERATIONS_TO_FUZZ = 100000

# Read MSR?
_READ_MSR = False

# Flush log file before each MSR
_FLUSH_LOG_EACH_MSR = False

# Control values to be written to each MSR
_FUZZ_VALUE_0_all1s = True
_FUZZ_VALUE_5A = False
_FUZZ_VALUE_RND = True

# Exclude MSRs which cause VM hang/crash
_EXCLUDE_MSR = []


class msr_fuzz (BaseModule):
    def __init__(self):
        BaseModule.__init__(self)

    def fuzz_MSRs(self, msr_addr_start, random_order=False):
        msr_addr_range = 0x10000
        msr_addr_end = msr_addr_start + msr_addr_range
        self.logger.log(f'[*] Fuzzing MSRs in range 0x{msr_addr_start:08X}:0x{msr_addr_end:08X}..')
        it = 0
        if random_order:
            it_max = _NO_ITERATIONS_TO_FUZZ
        else:
            it_max = msr_addr_range

        while it < it_max:
            if random_order:
                msr_addr = random.randint(msr_addr_start, msr_addr_end)
            else:
                msr_addr = msr_addr_start + it

            if _FLUSH_LOG_EACH_MSR:
                self.logger.flush()

            if msr_addr not in _EXCLUDE_MSR:
                if _READ_MSR:
                    self.logger.log(f'[*] rdmsr 0x{msr_addr:08X}')
                    try:
                        (_, _) = self.cs.msr.read_msr(0, msr_addr)
                    except:
                        pass

                self.logger.log(f'[*] wrmsr 0x{msr_addr:08X}')

                if _FUZZ_VALUE_0_all1s:
                    try:
                        self.cs.msr.write_msr(0, msr_addr, 0, 0)
                    except:
                        pass
                    try:
                        self.cs.msr.write_msr(0, msr_addr, 0xFFFFFFFF, 0xFFFFFFFF)
                    except:
                        pass

                if _FUZZ_VALUE_5A:
                    try:
                        self.cs.msr.write_msr(0, msr_addr, 0x5A5A5A5A, 0x5A5A5A5A)
                    except:
                        pass

                if _FUZZ_VALUE_RND:
                    val_hi = random.randint(0, 0xFFFFFFFF)
                    val_lo = random.randint(0, 0xFFFFFFFF)
                    try:
                        self.cs.msr.write_msr(0, msr_addr, val_hi, val_lo)
                    except:
                        pass
            it += 1
        return True

    def run(self, module_argv):
        self.logger.start_test('Fuzzing CPU Model Specific Registers (MSR)')

        _random_order = False
        if (len(module_argv) > 0) and ('random' == module_argv[0].lower()):
            _random_order = True

        global _NO_ITERATIONS_TO_FUZZ
        _NO_ITERATIONS_TO_FUZZ = 100000

        self.logger.log('[*] Configuration:')
        self.logger.log(f'    Mode: {"random" if _random_order else "sequential"}')
        if _random_order:
            self.logger.log(f'    Number of iterations: {_NO_ITERATIONS_TO_FUZZ:d}')

        self.logger.log('\n[*] Fuzzing Low MSR range...')
        self.fuzz_MSRs(0x0, _random_order)
        self.logger.log('\n[*] Fuzzing High MSR range...')
        self.fuzz_MSRs(0xC0000000, _random_order)
        self.logger.log('\n[*] Fuzzing VMM synthetic MSR range...')
        self.fuzz_MSRs(0x40000000, _random_order)

        self.logger.log_information('Module completed')
        self.logger.log_warning('System may be in an unknown state, further evaluation may be needed.')
        self.result.setStatusBit(self.result.status.VERIFY)
        self.res = self.result.getReturnCode(ModuleResult.WARNING)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util spidesc <rom>

Examples:

>>> chipsec_util spidesc spi.bin
"""

from chipsec.command import BaseCommand, toLoad
from chipsec.library.file import read_file
from chipsec.hal.spi_descriptor import parse_spi_flash_descriptor
from argparse import ArgumentParser


class SPIDescCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.Config

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util spidesc', usage=__doc__)
        parser.add_argument('fd_file', type=str, help='File name')
        parser.set_defaults()
        parser.parse_args(self.argv, namespace=self)

    def run(self):

        self.logger.log("[CHIPSEC] Parsing SPI Flash Descriptor from file '{}'\n".format(self.fd_file))
        fd = read_file(self.fd_file)
        if fd:
            parse_spi_flash_descriptor(self.cs, fd)


commands = {'spidesc': SPIDescCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util msgbus read     <port> <register>
>>> chipsec_util msgbus write    <port> <register> <value>
>>> chipsec_util msgbus mm_read  <port> <register>
>>> chipsec_util msgbus mm_write <port> <register> <value>
>>> chipsec_util msgbus message  <port> <register> <opcode> [value]
>>>
>>> <port>    : message bus port of the target unit
>>> <register>: message bus register/offset in the target unit port
>>> <value>   : value to be written to the message bus register/offset
>>> <opcode>  : opcode of the message on the message bus

Examples:

>>> chipsec_util msgbus read     0x3 0x2E
>>> chipsec_util msgbus mm_write 0x3 0x27 0xE0000001
>>> chipsec_util msgbus message  0x3 0x2E 0x10
>>> chipsec_util msgbus message  0x3 0x2E 0x11 0x0
"""

from chipsec.command import BaseCommand, toLoad
from argparse import ArgumentParser


# Message Bus
class MsgBusCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util msgbus', usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_read = subparsers.add_parser('read')
        parser_read.add_argument('port', type=lambda x: int(x, 16), help='Port (hex)')
        parser_read.add_argument('reg', type=lambda x: int(x, 16), help='Register (hex)')
        parser_read.set_defaults(func=self.msgbus_read)

        parser_write = subparsers.add_parser('write')
        parser_write.add_argument('port', type=lambda x: int(x, 16), help='Port (hex)')
        parser_write.add_argument('reg', type=lambda x: int(x, 16), help='Register (hex)')
        parser_write.add_argument('val', type=lambda x: int(x, 16), help='Value (hex)')
        parser_write.set_defaults(func=self.msgbus_write)

        parser_mmread = subparsers.add_parser('mm_read')
        parser_mmread.add_argument('port', type=lambda x: int(x, 16), help='Port (hex)')
        parser_mmread.add_argument('reg', type=lambda x: int(x, 16), help='Register (hex)')
        parser_mmread.set_defaults(func=self.msgbus_mm_read)

        parser_mmwrite = subparsers.add_parser('mm_write')
        parser_mmwrite.add_argument('port', type=lambda x: int(x, 16), help='Port (hex)')
        parser_mmwrite.add_argument('reg', type=lambda x: int(x, 16), help='Register (hex)')
        parser_mmwrite.add_argument('val', type=lambda x: int(x, 16), help='Value (hex)')
        parser_mmwrite.set_defaults(func=self.msgbus_mm_write)

        parser_message = subparsers.add_parser('message')
        parser_message.add_argument('port', type=lambda x: int(x, 16), help='Port (hex)')
        parser_message.add_argument('reg', type=lambda x: int(x, 16), help='Register (hex)')
        parser_message.add_argument('opcode', type=lambda x: int(x, 16), help='OPCODE (hex)')
        parser_message.add_argument('val', type=lambda x: int(x, 16), nargs='?', default=None, help='Value (hex)')
        parser_message.set_defaults(func=self.msgbus_message)

        parser.parse_args(self.argv, namespace=self)

    def msgbus_read(self):
        self.logger.log(f'[CHIPSEC] msgbus read: port 0x{self.port:02X} + 0x{self.reg:08X}')
        return self._msgbus.msgbus_reg_read(self.port, self.reg)

    def msgbus_write(self):
        self.logger.log(f'[CHIPSEC] msgbus write: port 0x{self.port:02X} + 0x{self.reg:08X} < 0x{self.val:08X}')
        return self._msgbus.msgbus_reg_write(self.port, self.reg, self.val)

    def msgbus_mm_read(self):
        self.logger.log(f'[CHIPSEC] MMIO msgbus read: port 0x{self.port:02X} + 0x{self.reg:08X}')
        return self._msgbus.mm_msgbus_reg_read(self.port, self.reg)

    def msgbus_mm_write(self):
        self.logger.log(f'[CHIPSEC] MMIO msgbus write: port 0x{self.port:02X} + 0x{self.reg:08X} < 0x{self.val:08X}')
        return self._msgbus.mm_msgbus_reg_write(self.port, self.reg, self.val)

    def msgbus_message(self):
        self.logger.log(f'[CHIPSEC] msgbus message: port 0x{self.port:02X} + 0x{self.reg:08X}, opcode: 0x{self.opcode:02X}')
        if self.val is not None:
            self.logger.log(f'[CHIPSEC]                 Data: 0x{self.val:08X}')
        return self._msgbus.msgbus_send_message(self.port, self.reg, self.opcode, self.val)

    def run(self):
        self._msgbus = self.cs.msgbus

        res = self.func()

        if res is not None:
            self.logger.log(f'[CHIPSEC] Result: {hex(res)}')


commands = {'msgbus': MsgBusCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Hyper-V VMBus synthetic keyboard fuzzer. Fuzzes inbound ring buffer in VMBus virtual keyboard device.

Usage:
  ``chipsec_main.py -m tools.vmm.hv.synth_kbd -a fuzz -l log.txt``

Note: the fuzzer is incompatible with native VMBus driver (``vmbus.sys``). To use it, remove ``vmbus.sys``
"""

import sys
import traceback
from random import randint
from struct import pack, unpack
from chipsec.library.defines import DD
from chipsec.library.returncode import ModuleResult
from chipsec.module_common import BaseModule, CPU
from chipsec.modules.tools.vmm.common import overwrite, session_logger
from chipsec.modules.tools.vmm.hv.define import HV_KBD_GUID, VMBUS_DATA_PACKET_FLAG_COMPLETION_REQUESTED, vm_pkt
from chipsec.modules.tools.vmm.hv.vmbus import RingBuffer, VMBusDiscovery

SYNTH_KBD_VERSION = 0x00010000
SYNTH_KBD_PROTOCOL_REQUEST = 1
SYNTH_KBD_PROTOCOL_RESPONSE = 2
SYNTH_KBD_EVENT = 3
SYNTH_KBD_LED_INDICATORS = 4

sys.stdout = session_logger(True, 'synth_kbd')

TAGS = [CPU]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'TOOLS', 'VMM', 'HV', 'SYNTH_KBD']


class RingBufferFuzzer(RingBuffer):
    def __init__(self):
        RingBuffer.__init__(self)
        self.fuzzing = False
        self.count = 0

    ##
    # ringbuffer_read - Fuzzing recv ring buffer pointers
    ##
    def ringbuffer_read(self):
        if self.fuzzing:
            buffer = self.cs.mem.read_physical_mem(self.pfn[self.send_size], 0x10)
            write_index, read_index, interrupt_mask, pending_send_sz = unpack('<4L', buffer)
            overwrite(buffer, DD(randint(0, 0xFFFFFFFF)), 4 * randint(0, 3))
            self.cs.mem.write_physical_mem(self.pfn[self.send_size], len(buffer), buffer)
            result = ''
            self.count += 1
            if self.count > 1000000:
                raise Exception
        else:
            result = RingBuffer.ringbuffer_read(self)
        return result


class synth_kbd(BaseModule):
    def __init__(self):
        BaseModule.__init__(self)

    def usage(self):
        self.logger.log(__doc__)
        return

    def run(self, module_argv):
        self.logger.start_test("Hyper-V VMBus virtual keyboard fuzzer")

        if len(module_argv) > 0:
            command = module_argv[0]
        else:
            self.usage()
            self.result.setStatusBit(self.result.status.UNSUPPORTED_OPTION)
            return self.result.getReturnCode(ModuleResult.ERROR)

        vb = VMBusDiscovery()
        vb.debug = True
        vb.promt = 'VMBUS KBD'
        vb.vmbus_init()
        vb.vmbus_connect()
        vb.vmbus_request_offers()
        relid = vb.get_relid_by_guid(HV_KBD_GUID)
        if relid == 0:
            vb.fatal(f'Could not found keyboard device with GUID: {HV_KBD_GUID}')

        vb.ringbuffers[relid] = RingBufferFuzzer()
        vb.ringbuffers[relid].ringbuffer_alloc(4)
        vb.ringbuffers[relid].gpadl = vb.vmbus_get_next_gpadl()

        vb.vmbus_establish_gpadl(relid, vb.ringbuffers[relid].gpadl, vb.ringbuffers[relid].pfn)
        vb.vmbus_open(relid, vb.ringbuffers[relid].gpadl, vb.ringbuffers[relid].send_size)
        try:
            vb.print_offer_channels()
            vb.print_created_gpadl()
            vb.print_open_channels()

            synth_kbd_protocol_request = pack('<LL', SYNTH_KBD_PROTOCOL_REQUEST, SYNTH_KBD_VERSION)
            vmpkt_datainband = list(vm_pkt.keys())[list(vm_pkt.values()).index('VM_PKT_DATA_INBAND')]
            vb.vmbus_sendpacket(relid, synth_kbd_protocol_request, 0x0, vmpkt_datainband, VMBUS_DATA_PACKET_FLAG_COMPLETION_REQUESTED)
            synth_kbd_protocol_response = vb.vmbus_recvpacket(relid)
            if len(synth_kbd_protocol_response) != 8:
                vb.fatal('Invalid response from synthetic keyboard!')
            msg_type, proto_status = unpack('<LL', synth_kbd_protocol_response)
            if (proto_status & 0x1) == 0x1:
                vb.msg('synth_kbd protocol request has been accepted!')
                vb.ringbuffers[relid].debug = False
                vb.debug = False
                while True:
                    synth_kbd_msg = vb.vmbus_recvpacket(relid)
                    if not synth_kbd_msg:
                        continue
                    if len(synth_kbd_msg) < 12:
                        vb.hex('invalid message', synth_kbd_msg)
                        continue
                    msg_type, code, rsvd, info = unpack('<LHHL', synth_kbd_msg[:12])
                    if msg_type == SYNTH_KBD_EVENT:
                        vb.msg(f'keystroke: {code:04X}  flags: {info:08X}')
                        vb.ringbuffers[relid].fuzzing = (command == 'fuzzing')
                        if code == 0x0046:
                            vb.msg('*** Control Break ***')
                            vb.ringbuffers[relid].fuzzing = False
                            break
                    else:
                        vb.hex(f'unhandled message type: {msg_type:d}', synth_kbd_msg)
            else:
                vb.err('synth_kbd protocol request has failed!')

        except KeyboardInterrupt:
            self.logger.log('***** Control-C *****')
        except Exception:
            self.logger.log('\n\n')
            self.logger.log_bad(traceback.format_exc())
            self.logger.log('\n\n')

        finally:
            vb.vmbus_close(relid)
            vb.vmbus_teardown_gpadl(relid, vb.ringbuffers[relid].gpadl)
            vb.vmbus_rescind_all_offers()
            del vb.ringbuffers[relid]
            del vb
        self.result.setStatusBit(self.result.status.SUCCESS)
        return self.result.getReturnCode(ModuleResult.PASSED)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2020, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
Verify that all Secure Boot key UEFI variables are authenticated (BS+RT+AT)
and protected from unauthorized modification.

Reference:
    - `UEFI 2.4 spec Section 28 <http://uefi.org/>`_

Usage:
    ``chipsec_main -m common.secureboot.variables [-a modify]``
    - ``-a`` : modify = will try to write/corrupt the variables

Where:
    - ``[]``: optional line

Examples:
    >>> chipsec_main.py -m common.secureboot.variables
    >>> chipsec_main.py -m common.secureboot.variables -a modify

.. note::
    - Module is not supported in all environments.

"""


from chipsec.module_common import BaseModule, SECUREBOOT, OPT_MODIFY
from chipsec.library.returncode import ModuleResult
from chipsec.hal.uefi import UEFI, SECURE_BOOT_VARIABLES, IS_VARIABLE_ATTRIBUTE, EFI_VAR_NAME_SecureBoot, SECURE_BOOT_KEY_VARIABLES
from chipsec.hal.uefi import EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS, EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS
from chipsec.hal.uefi import SECURE_BOOT_OPTIONAL_VARIABLES
from chipsec.hal.uefi_common import StatusCode
from typing import AnyStr, List, Optional

_MODULE_NAME = 'variables'


TAGS = [SECUREBOOT]
METADATA_TAGS = ['OPENSOURCE', 'IA', 'COMMON', 'SECUREBOOT', 'VARIABLES']


class variables(BaseModule):

    def __init__(self):
        BaseModule.__init__(self)
        self._uefi = UEFI(self.cs)
        self.result.url = 'https://chipsec.github.io/modules/chipsec.modules.common.secureboot.variables.html'

    def is_supported(self) -> bool:
        supported = self.cs.helper.EFI_supported()
        if not supported:
            self.logger.log_important('OS does not support UEFI Runtime API.  Skipping module.')
        return supported

    def can_modify(self, name: str, guid: Optional[AnyStr], data: Optional[bytes]) -> bool:
        if not guid or not data:
            self.logger.log(f'    > Missing GUID or Data. Unable to modify variable {guid}:{name} data:{data}')
            return False
        else:
            self.logger.log(f'    > Attempting to modify variable {guid}:{name}')

        baddata = (data[0] ^ 0xFF).to_bytes(1, 'little') + data[1:]
        status = self._uefi.set_EFI_variable(name, guid, baddata)
        if StatusCode.EFI_SUCCESS != status:
            self.logger.log(f'    < Modification of {name} returned error 0x{status:X}')
        else:
            self.logger.log(f'    < Modification of {name} returned success')

        self.logger.log(f'    > Checking variable {name} contents after modification..')
        newdata = self._uefi.get_EFI_variable(name, guid)

        _changed = data != newdata
        if _changed:
            self.logger.log_bad(f'EFI variable {name} has been modified. Restoring original contents..')
            self._uefi.set_EFI_variable(name, guid, data)

            restoreddata = self._uefi.get_EFI_variable(name, guid)
            if (restoreddata != data):
                self.logger.log_important(f'Failed to restore contents of variable {name} failed!')
            else:
                self.logger.log(f'    Contents of variable {name} have been restored')
        else:
            self.logger.log_good(f'Could not modify UEFI variable {guid}:{name}')
        return _changed

    def check_secureboot_variable_attributes(self, do_modify: bool) -> int:
        not_found = 0
        not_auth = 0
        not_wp = 0
        is_secureboot_enabled = False

        sbvars = self._uefi.list_EFI_variables()
        if sbvars is None:
            self.logger.log_warning('Could not enumerate UEFI variables.')
            self.result.setStatusBit(self.result.status.CONFIGURATION)
            return self.result.getReturnCode(ModuleResult.WARNING)

        for name in SECURE_BOOT_VARIABLES:

            if (name in sbvars.keys()) and (sbvars[name] is not None):
                if len(sbvars[name]) > 1:
                    self.logger.log_failed(f'There should only be one instance of variable {name}')
                    self.result.setStatusBit(self.result.status.VERIFY)
                    return self.result.getReturnCode(ModuleResult.FAILED)
                for (_, _, _, data, guid, attrs) in sbvars[name]:
                    self.logger.log(f'[*] Checking protections of UEFI variable {guid}:{name}')

                    if EFI_VAR_NAME_SecureBoot == name:
                        is_secureboot_enabled = (data is not None) and (len(data) == 1) and (ord(data) == 0x1)

                    if name in SECURE_BOOT_KEY_VARIABLES:
                        if IS_VARIABLE_ATTRIBUTE(attrs, EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS):
                            self.logger.log_good(f'Variable {guid}:{name} is authenticated (AUTHENTICATED_WRITE_ACCESS)')
                        elif IS_VARIABLE_ATTRIBUTE(attrs, EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS):
                            self.logger.log_good(f'Variable {guid}:{name} is authenticated (TIME_BASED_AUTHENTICATED_WRITE_ACCESS)')
                        else:
                            not_auth += 1
                            self.logger.log_bad(f'Variable {guid}:{name} is not authenticated')

                    if do_modify:
                        if self.can_modify(name, guid, data):
                            not_wp += 1
            elif name in SECURE_BOOT_OPTIONAL_VARIABLES:
                self.logger.log_important(f'Secure Boot variable {name} is not found but is optional')
                continue
            else:
                not_found += 1
                self.logger.log_important(f'Secure Boot variable {name} is not found')
                continue

        self.logger.log('')
        prefix = 'en' if is_secureboot_enabled else 'dis'
        self.logger.log(f'[*] Secure Boot appears to be {prefix}abled')

        if len(SECURE_BOOT_VARIABLES) == not_found:
            self.logger.log_warning('None of required Secure Boot variables found.')
            self.logger.log_important('If Secure Boot is enabled, this could be a problem.')
            self.result.setStatusBit(self.result.status.VERIFY)
            return self.result.getReturnCode(ModuleResult.WARNING)
        else:
            sb_vars_failed = (not_found > 0) or (not_auth > 0) or (not_wp > 0)
            if sb_vars_failed:
                if not_found > 0:
                    self.logger.log_bad('Some required Secure Boot variables are missing')
                if not_auth > 0:
                    self.logger.log_bad('Some Secure Boot keying variables are not authenticated')
                if not_wp > 0:
                    self.logger.log_bad('Some Secure Boot variables can be modified')

                if is_secureboot_enabled:
                    self.logger.log_failed('Not all Secure Boot UEFI variables are protected')
                    self.result.setStatusBit(self.result.status.PROTECTION)
                    return self.result.getReturnCode(ModuleResult.FAILED)
                else:
                    self.logger.log_warning('Not all Secure Boot UEFI variables are protected')
                    self.result.setStatusBit(self.result.status.FEATURE_DISABLED)
                    return self.result.getReturnCode(ModuleResult.WARNING)

            else:
                self.logger.log_passed('All Secure Boot UEFI variables are protected')
                self.result.setStatusBit(self.result.status.SUCCESS)
                return self.result.getReturnCode(ModuleResult.PASSED)

    def run(self, module_argv: List[str]) -> int:
        self.logger.start_test('Attributes of Secure Boot EFI Variables')

        do_modify = (len(module_argv) > 0) and (module_argv[0] == OPT_MODIFY)

        self.res = self.check_secureboot_variable_attributes(do_modify)
        return self.res



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
The igd command allows memory read/write operations using igd dma.

>>> chipsec_util igd
>>> chipsec_util igd dmaread <address> [width] [file_name]
>>> chipsec_util igd dmawrite <address> <width> <value|file_name>

Examples:

>>> chipsec_util igd dmaread 0x20000000 4
>>> chipsec_util igd dmawrite 0x2217F1000 0x4 deadbeef
"""

from chipsec.command import BaseCommand, toLoad
from chipsec.library.logger import print_buffer_bytes
from argparse import ArgumentParser
from chipsec.library.file import read_file, write_file
import os


# Port I/O
class IgdCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util igd', usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_read = subparsers.add_parser('dmaread')
        parser_read.add_argument('address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_read.add_argument('width', type=lambda x: int(x, 16), nargs='?', default=0x100, help='Width of read (hex)')
        parser_read.add_argument('file_name', type=str, nargs='?', default='', help='File name to save data')
        parser_read.set_defaults(func=self.read_dma)

        parser_write = subparsers.add_parser('dmawrite')
        parser_write.add_argument('address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_write.add_argument('size', type=lambda x: int(x, 16), help='Size of data to write (hex)')
        parser_write.add_argument('file_value', type=str, help='Data to write [Value|<file_name>]')
        parser_write.set_defaults(func=self.write_dma)

        parser.parse_args(self.argv, namespace=self)

    def read_dma(self) -> None:
        self.logger.log(f'[CHIPSEC] Reading buffer from memory: PA = 0x{self.address:016X}, len = 0x{self.width:X}..')
        buffer = self.cs.igd.gfx_aperture_dma_read_write(self.address, self.width)
        if self.file_name:
            write_file(self.file_name, buffer)
            self.logger.log(f'[CHIPSEC] Written 0x{len(buffer):X} bytes to \'{self.file_name}\'')
        else:
            print_buffer_bytes(buffer)

    def write_dma(self) -> None:
        if not os.path.exists(self.file_value):
            buffer_value = self.file_value.lower().strip('0x')
            try:
                buffer = bytearray.fromhex(buffer_value)
            except ValueError as e:
                self.logger.log_error(f'Incorrect <value> specified: \'{self.file_value}\'')
                self.logger.log_error(str(e))
                return
            self.logger.log(f'[CHIPSEC] Read 0x{len(buffer):X} hex bytes from command-line: \'{buffer_value}\'')
        else:
            buffer = read_file(self.file_value)
            self.logger.log(f'[CHIPSEC] Read 0x{len(buffer):X} bytes from file \'{self.file_value}\'')

        if len(buffer) < self.size:
            self.logger.log_error(f'Number of bytes read (0x{len(buffer):X}) is less than the specified <length> (0x{self.size:X})')
            return

        self.logger.log(f'[CHIPSEC] Writing buffer to memory: PA = 0x{self.address:016X}, len = 0x{self.size:X}..')
        self.cs.igd.gfx_aperture_dma_read_write(self.address, self.size, buffer)

    def run(self) -> None:

        if not self.cs.igd.is_device_enabled():
            self.logger.log('[CHIPSEC] Looks like internal graphics device is not enabled')
            return

        self.func()


commands = {'igd': IgdCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Platform specific UEFI functionality (parsing platform specific EFI NVRAM, capsules, etc.)
"""

import struct
from collections import namedtuple
from uuid import UUID
from typing import Dict, List, Tuple, Optional, Union, Any
from chipsec.library import defines
from chipsec.library.logger import logger
from chipsec.hal.uefi_common import bit_set, VARIABLE_SIGNATURE_VSS, S3BootScriptOpcode_MDE, op_io_pci_mem, S3BootScriptOpcode_EdkCompat, EFI_GUID_STR, EFI_GUID_SIZE
from chipsec.hal.uefi_common import op_stall, op_dispatch, op_terminate, op_mem_poll, op_unknown, get_3b_size, get_nvar_name, op_smbus_execute, script_width_formats
from chipsec.hal.uefi_common import S3BOOTSCRIPT_ENTRY, MAX_S3_BOOTSCRIPT_ENTRY_LENGTH, VARIABLE_STORE_FV_GUID, IS_VARIABLE_ATTRIBUTE, VARIABLE_DATA
from chipsec.hal.uefi_common import EFI_VARIABLE_BOOTSERVICE_ACCESS, EFI_VARIABLE_NON_VOLATILE, EFI_VARIABLE_RUNTIME_ACCESS, script_opcodes
from chipsec.hal.uefi_common import EFI_VARIABLE_HARDWARE_ERROR_RECORD, EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS, EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS
from chipsec.hal.uefi_fv import NextFwVolume, NextFwFile, EFI_FVB2_ERASE_POLARITY, EFI_FV_FILETYPE_RAW

EfiTableType = Union['EFI_HDR_VSS', 'EFI_HDR_VSS_AUTH', 'EFI_HDR_VSS_APPLE', None]
EfiVariableType = Tuple[int, bytes, EfiTableType, bytes, str, int]

#
# List of supported types of EFI NVRAM format (platform/vendor specific)
#


class FWType:
    EFI_FW_TYPE_UEFI = 'uefi'
    EFI_FW_TYPE_UEFI_AUTH = 'uefi_auth'
#    EFI_FW_TYPE_WIN       = 'win'      # Windows 8 GetFirmwareEnvironmentVariable format
    EFI_FW_TYPE_VSS = 'vss'       # NVRAM using format with '$VSS' signature
    EFI_FW_TYPE_VSS_AUTH = 'vss_auth'  # NVRAM using format with '$VSS' signature with extra fields
    # See "A Tour Beyond BIOS Implementing UEFI Authenticated
    # Variables in SMM with EDKII"
    EFI_FW_TYPE_VSS2 = 'vss2'
    EFI_FW_TYPE_VSS2_AUTH = 'vss2_auth'
    EFI_FW_TYPE_VSS_APPLE = 'vss_apple'
    EFI_FW_TYPE_NVAR = 'nvar'      # 'NVAR' NVRAM format
    EFI_FW_TYPE_EVSA = 'evsa'      # 'EVSA' NVRAM format


fw_types: List[str] = []
for i in [t for t in dir(FWType) if not callable(getattr(FWType, t))]:
    if not i.startswith('__'):
        fw_types.append(getattr(FWType, i))


NVRAM_ATTR_RT = 1
NVRAM_ATTR_DESC_ASCII = 2
NVRAM_ATTR_GUID = 4
NVRAM_ATTR_DATA = 8
NVRAM_ATTR_EXTHDR = 0x10
NVRAM_ATTR_AUTHWR = 0x40
NVRAM_ATTR_HER = 0x20
NVRAM_ATTR_VLD = 0x80

#
# Known GUIDs of NVRAM stored in EFI firmware volumes, FS files etc. of various firmware implementations
#
ADDITIONAL_NV_STORE_GUID = UUID('00504624-8A59-4EEB-BD0F-6B36E96128E0')
NVAR_NVRAM_FS_FILE = UUID("CEF5B9A3-476D-497F-9FDC-E98143E0422C")

LENOVO_FS1_GUID = UUID("16B45DA2-7D70-4AEA-A58D-760E9ECB841D")
LENOVO_FS2_GUID = UUID("E360BDBA-C3CE-46BE-8F37-B231E5CB9F35")

EFI_PLATFORM_FS_GUIDS = [LENOVO_FS1_GUID, LENOVO_FS2_GUID]
EFI_NVRAM_GUIDS = [VARIABLE_STORE_FV_GUID, ADDITIONAL_NV_STORE_GUID, NVAR_NVRAM_FS_FILE]

#
# This Variable header is defined by UEFI
#

#
# Variable Store Status
#
# typedef enum {
#  EfiRaw,
#  EfiValid,
#  EfiInvalid,
#  EfiUnknown
# } VARIABLE_STORE_STATUS;
VARIABLE_STORE_STATUS_RAW = 0
VARIABLE_STORE_STATUS_VALID = 1
VARIABLE_STORE_STATUS_INVALID = 2
VARIABLE_STORE_STATUS_UNKNOWN = 3

#
# Variable State flags
#
VAR_IN_DELETED_TRANSITION = 0xfe  # Variable is in obsolete transition
VAR_DELETED = 0xfd  # Variable is obsolete
VAR_ADDED = 0x7f  # Variable has been completely added


def IS_VARIABLE_STATE(_c: int, _Mask: int) -> bool:
    return ((((~_c) & 0xFF) & ((~_Mask) & 0xFF)) != 0)


#
# typedef struct {
#  UINT16    StartId;
#  UINT8     State;
#  UINT8     Reserved;
#  UINT32    Attributes;
#  UINT32    NameSize;
#  UINT32    DataSize;
#  EFI_GUID  VendorGuid;
# } VARIABLE_HEADER;
#
# typedef struct {
#  UINT32  Data1;
#  UINT16  Data2;
#  UINT16  Data3;
#  UINT8   Data4[8];
# } EFI_GUID;
#
UEFI_VARIABLE_HEADER_SIZE = 28


class UEFI_VARIABLE_HEADER(namedtuple('UEFI_VARIABLE_HEADER', 'StartId State Reserved Attributes NameSize DataSize VendorGuid0 VendorGuid1 VendorGuid2 VendorGuid3')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
Header (UEFI)
-------------
StartId    : 0x{self.StartId:04X}
State      : 0x{self.State:02X}
Reserved   : 0x{self.Reserved:02X}
Attributes : 0x{self.Attributes:08X}
NameSize   : 0x{self.NameSize:08X}
DataSize   : 0x{self.DataSize:08X}
VendorGuid : {{0x{self.VendorGuid0:08X}-0x{self.VendorGuid1:04X}-0x{self.VendorGuid2:04X}-0x{self.VendorGuid3:08X}}}
"""


UEFI_VARIABLE_STORE_HEADER = "<16sIBBHI"
UEFI_VARIABLE_STORE_HEADER_SIZE = struct.calcsize(UEFI_VARIABLE_STORE_HEADER)
'''
EFI_VARIABLE_HEADER_AUTH = "<HBBI28sIIIHH8s"
EFI_VARIABLE_HEADER_AUTH_SIZE = struct.calcsize(EFI_VARIABLE_HEADER_AUTH)

EFI_VARIABLE_HEADER = "<HBBIIIIHH8s"
EFI_VARIABLE_HEADER_SIZE = struct.calcsize(EFI_VARIABLE_HEADER)
'''
VARIABLE_STORE_FORMATTED = 0x5a
VARIABLE_STORE_HEALTHY = 0xfe

NvStore = Tuple[int, int, None]


def _getNVstore_EFI(nvram_buf: bytes, efi_type: str) -> NvStore:
    l = (-1, -1, None)
    FvOffset = 0
    FvLength = 0
    fv = NextFwVolume(nvram_buf, FvOffset, FvLength)
    while True:
        if (fv is None):
            break
        if (fv.Guid == VARIABLE_STORE_FV_GUID):
            nvram_start = fv.HeaderSize
            _, _, Format, State, _, _ = struct.unpack(UEFI_VARIABLE_STORE_HEADER, fv.Image[nvram_start:nvram_start + UEFI_VARIABLE_STORE_HEADER_SIZE])
            if ((Format == VARIABLE_STORE_FORMATTED) and (State == VARIABLE_STORE_HEALTHY)):
                if (isCorrectVSStype(fv.Image[nvram_start:], efi_type)):
                    l = (fv.Offset + nvram_start, fv.Size - nvram_start, None)
                break
        fv = NextFwVolume(nvram_buf, fv.Offset, fv.Size)
    return l


def getNVstore_EFI(nvram_buf: bytes) -> NvStore:
    return _getNVstore_EFI(nvram_buf, FWType.EFI_FW_TYPE_VSS)


def getNVstore_EFI_AUTH(nvram_buf: bytes) -> NvStore:
    return _getNVstore_EFI(nvram_buf, FWType.EFI_FW_TYPE_VSS_AUTH)


def getEFIvariables_UEFI(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS)


def getEFIvariables_UEFI_AUTH(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS_AUTH)


'''
def getEFIvariables_UEFI_Ex( nvram_buf, auth = False ):
    dof = 0
    length = len(nvram_buf)
    storen = 0
    variables = dict()
    while ((dof+UEFI_VARIABLE_STORE_HEADER_SIZE) < length):
        store_start = dof
        StoreGuid0, StoreGuid1, StoreGuid2, StoreGuid03, Size, Format, State, R0, R1 = \
            struct.unpack(UEFI_VARIABLE_STORE_HEADER, nvram_buf[dof:dof + UEFI_VARIABLE_STORE_HEADER_SIZE])
        dof = align(dof + UEFI_VARIABLE_STORE_HEADER_SIZE, 4)
        if ((Format != VARIABLE_STORE_FORMATTED) or (State != VARIABLE_STORE_HEALTHY)):
            break
        if ((store_start + Size) >= length): break
        while ((dof + EFI_VARIABLE_HEADER_SIZE) <= (store_start + Size)):
            StartId, State, R0, Attributes, Auth, NameSize, DataSize, VendorGuid0, VendorGuid1, VendorGuid2, VendorGuid3 = \
                struct.unpack(EFI_VARIABLE_HEADER, nvram_buf[dof:dof+EFI_VARIABLE_HEADER_SIZE]);
            if (StartId != VARIABLE_DATA): break
            dof += EFI_VARIABLE_HEADER_SIZE
            if ((State == 0xff) and (DataSize == 0xffffffff) and (NameSize == 0xffffffff) and (Attributes == 0xffffffff)):
                NameSize = 0
                DataSize = 0
                # just skip variable with empty name and data for now
            else:
                guid = guid_str(VendorGuid0, VendorGuid1, VendorGuid2, VendorGuid3)
                Name = nvram_buf[dof:dof+NameSize]
                NameStr = unicode(Name, "utf-16-le").split('\x00')[0]
                VarData = nvram_buf[dof+NameSize:dof+NameSize+DataSize]
                if NameStr not in variables.keys():
                    variables[NameStr] = []
                #                          off, buf,  hdr,  data,    guid, attrs
                variables[NameStr].append((dof, None, None, VarData, guid, Attributes))
            dof = align(dof+NameSize+DataSize, 4)
        dof = store_start + Size
        storen += 1
    return variables
'''
##################################################################################################
#
# Platform/Vendor Specific EFI NVRAM Parsing Functions
#
# For each platform, EFI NVRAM parsing functionality includes:
# 1. Function to parse EFI variable within NVRAM binary (func_getefivariables)
#    May define/use platform specific EFI Variable Header
#    Function arguments:
#      In : binary buffer (as a string)
#      Out:
#        start           - offset in the buffer to the current EFI variable
#        next_var_offset - offset in the buffer to the next EFI variable
#        efi_var_buf     - full EFI variable buffer
#        efi_var_hdr     - EFI variable header object
#        efi_var_name    - EFI variable name
#        efi_var_data    - EFI variable data contents
#        efi_var_guid    - EFI variable GUID
#        efi_var_attr    - EFI variable attributes
# 2. [Optional] Function to find EFI NVRAM within arbitrary binary (func_getnvstore)
#    If this function is not defined, 'chipsec_util uefi' searches EFI variables from the beginning of the binary
#    Function arguments:
#      In : NVRAM binary buffer (as a string)
#      Out:
#        start        - offset of NVRAM     (-1 means NVRAM not found)
#        size         - size of NVRAM       (-1 means NVRAM is entire binary)
#        nvram_header - NVRAM header object
#
##################################################################################################

##################################################################################################
# NVAR format of NVRAM
#


class EFI_HDR_NVAR1(namedtuple('EFI_HDR_NVAR1', 'StartId TotalSize Reserved1 Reserved2 Reserved3 Attributes State')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
Header (NVAR)
------------
StartId    : 0x{self.StartId:04X}
TotalSize  : 0x{self.TotalSize:04X}
Reserved1  : 0x{self.Reserved1:02X}
Reserved2  : 0x{self.Reserved2:02X}
Reserved3  : 0x{self.Reserved3:02X}
Attributes : 0x{self.Attributes:02X}
State      : 0x{self.State:02X}
"""


NVAR_EFIvar_signature = b'NVAR'


def getNVstore_NVAR(nvram_buf: bytes) -> NvStore:
    l = (-1, -1, None)
    fv = NextFwVolume(nvram_buf)
    if (fv is None):
        return l
    if (fv.Offset >= len(nvram_buf)):
        return l
    if (fv.Offset + fv.Size) > len(nvram_buf):
        fv.Size = len(nvram_buf) - fv.Offset
    while fv is not None:
        polarity = bit_set(fv.Attributes, EFI_FVB2_ERASE_POLARITY)
        fwbin = NextFwFile(fv.Image, fv.Size, fv.HeaderSize, polarity)
        while fwbin is not None:
            if (fwbin.Type == EFI_FV_FILETYPE_RAW) and (fwbin.Guid == NVAR_NVRAM_FS_FILE):
                l = ((fv.Offset + fwbin.Offset + fwbin.HeaderSize), fwbin.Size - fwbin.HeaderSize, None)
                if (not fwbin.UD):
                    return l
            fwbin = NextFwFile(fv.Image, fv.Size, fwbin.Size + fwbin.Offset, polarity)
        fv = NextFwVolume(nvram_buf, fv.Offset, fv.Size)
    return l


def _ord(c: Union[str, int]) -> int:
    return ord(c) if isinstance(c, str) else c


def getEFIvariables_NVAR(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    name = ''
    start = nvram_buf.find(NVAR_EFIvar_signature)
    nvram_size = len(nvram_buf)
    EFI_HDR_NVAR = "<4sH3sB"
    nvar_size = struct.calcsize(EFI_HDR_NVAR)
    variables = dict()
    nof = 0  # start
#   EMPTY = 0
    EMPTY = 0xffffffff
    while (nof + nvar_size) < nvram_size:
        start_id, size, next, attributes = struct.unpack(EFI_HDR_NVAR, nvram_buf[nof:nof + nvar_size])
        if size == 0:
            break
        next = get_3b_size(next)
        valid = (bit_set(attributes, NVRAM_ATTR_VLD) and (not bit_set(attributes, NVRAM_ATTR_DATA)))
        if not valid:
            nof = nof + size
            continue
        isvar = (start_id == NVAR_EFIvar_signature)
        if (not isvar) or (size == (EMPTY & 0xffff)):
            break
        var_name_off = 1
        if bit_set(attributes, NVRAM_ATTR_GUID):
            guid = UUID(bytes_le=nvram_buf[nof + nvar_size: nof + nvar_size + EFI_GUID_SIZE])
            guid = str(guid).upper()
            var_name_off = EFI_GUID_SIZE
        else:
            guid_idx = _ord(nvram_buf[nof + nvar_size])
            guid_off = (nvram_size - EFI_GUID_SIZE) - guid_idx * EFI_GUID_SIZE
            guid = UUID(bytes_le=nvram_buf[guid_off: guid_off + EFI_GUID_SIZE])
            guid = str(guid).upper()
        name_size = 0
        name_offset = nof + nvar_size + var_name_off
        if not bit_set(attributes, NVRAM_ATTR_DATA):
            name, name_size = get_nvar_name(nvram_buf, name_offset, bit_set(attributes, NVRAM_ATTR_DESC_ASCII))
        esize = 0
        eattrs = 0
        if bit_set(attributes, NVRAM_ATTR_EXTHDR):
            esize, = struct.unpack("<H", nvram_buf[nof + size - 2:nof + size])
            eattrs = _ord(nvram_buf[nof + size - esize])
        attribs = EFI_VARIABLE_BOOTSERVICE_ACCESS
        attribs = attribs | EFI_VARIABLE_NON_VOLATILE
        if bit_set(attributes, NVRAM_ATTR_RT):
            attribs = attribs | EFI_VARIABLE_RUNTIME_ACCESS
        if bit_set(attributes, NVRAM_ATTR_HER):
            attribs = attribs | EFI_VARIABLE_HARDWARE_ERROR_RECORD
        if bit_set(attributes, NVRAM_ATTR_AUTHWR):
            if bit_set(eattrs, EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS):
                attribs = attribs | EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS
            if bit_set(eattrs, EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS):
                attribs = attribs | EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS
        # Get variable data
        lof = nof
        lnext = next
        lattributes = attributes
        lsize = size
        lesize = esize
        while lnext != (0xFFFFFF & EMPTY):
            lof = lof + lnext
            lstart_id, lsize, lnext, lattributes = struct.unpack(EFI_HDR_NVAR, nvram_buf[lof:lof + nvar_size])
            lnext = get_3b_size(lnext)
        dataof = lof + nvar_size
        if not bit_set(lattributes, NVRAM_ATTR_DATA):
            lnameof = 1
            if bit_set(lattributes, NVRAM_ATTR_GUID):
                lnameof = EFI_GUID_SIZE
            name_offset = lof + nvar_size + lnameof
            name, name_size = get_nvar_name(nvram_buf, name_offset, bit_set(attributes, NVRAM_ATTR_DESC_ASCII))
            dataof = name_offset + name_size
        if bit_set(lattributes, NVRAM_ATTR_EXTHDR):
            lesize, = struct.unpack("<H", nvram_buf[lof + lsize - 2:lof + lsize])
        data = nvram_buf[dataof:lof + lsize - lesize]
        if name not in variables.keys():
            variables[name] = []
        #                       off, buf,  hdr,  data, guid, attrs
        variables[name].append((nof, b'', None, data, guid, attribs))
        nof = nof + size
    return variables


NVAR_HDR_FMT = '=IHBBBBB'
NVAR_HDR_SIZE = struct.calcsize(NVAR_HDR_FMT)


#
# Linear/simple NVAR format parsing
#
def getNVstore_NVAR_simple(nvram_buf: bytes) -> Tuple[Optional[int], int, None]:
    return (nvram_buf.find(NVAR_EFIvar_signature), -1, None)


def getEFIvariables_NVAR_simple(nvram_buf: bytes) -> Dict[str, Tuple[int, bytes, bytes, int, str, int]]:
    nvsize = len(nvram_buf)
    hdr_fmt = NVAR_HDR_FMT
    hdr_size = struct.calcsize(hdr_fmt)
    variables = dict()
    start = nvram_buf.find(NVAR_EFIvar_signature)
    if -1 == start:
        return variables

    while (start + hdr_size) < nvsize:
        efi_var_hdr = EFI_HDR_NVAR1(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))
        name_size = 0
        efi_var_name = "NA"
        if not IS_VARIABLE_ATTRIBUTE(efi_var_hdr.Attributes, EFI_VARIABLE_HARDWARE_ERROR_RECORD):
            name_size = nvram_buf[start + hdr_size:].find(b'\x00')
            efi_var_name = nvram_buf[start + hdr_size: start + hdr_size + name_size].decode('latin1')

        next_var_offset = start + efi_var_hdr.TotalSize
        data_size = efi_var_hdr.TotalSize - name_size - hdr_size
        efi_var_buf = nvram_buf[start: next_var_offset]
        efi_var_data = nvram_buf[start + hdr_size + name_size: next_var_offset]

        if efi_var_name not in variables.keys():
            variables[efi_var_name] = []
        #                               off,   buf,         hdr,         data,         guid, attrs
        variables[efi_var_name].append((start, efi_var_buf, efi_var_hdr, efi_var_data, '', efi_var_hdr.Attributes))

        if start >= next_var_offset:
            break
        start = next_var_offset

    return variables


#######################################################################
#
# VSS NVRAM (signature = '$VSS')
#
#

# define VARIABLE_STORE_SIGNATURE  EFI_SIGNATURE_32 ('$', 'V', 'S', 'S')
VARIABLE_STORE_SIGNATURE_VSS = b'$VSS'
VARIABLE_STORE_HEADER_FMT_VSS = '=IIBBHI'  # Signature is '$VSS'


class VARIABLE_STORE_HEADER_VSS(namedtuple('VARIABLE_STORE_HEADER_VSS', 'Signature Size Format State Reserved Reserved1')):
    __slots__ = ()

    def __str__(self) -> str:
        sig_str = struct.pack('=I', self.Signature)
        return f"""
EFI Variable Store
-----------------------------
Signature : {sig_str} (0x{self.Signature:08X})
Size      : 0x{self.Size:08X} bytes
Format    : 0x{self.Format:02X}
State     : 0x{self.State:02X}
Reserved  : 0x{self.Reserved:04X}
Reserved1 : 0x{self.Reserved1:08X}
"""


VARIABLE_STORE_SIGNATURE_VSS2 = UUID('DDCF3617-3275-4164-98B6-FE85707FFE7D').bytes_le
VARIABLE_STORE_SIGNATURE_VSS2_AUTH = UUID('AAF32C78-947B-439A-A180-2E144EC37792').bytes_le

VARIABLE_STORE_HEADER_FMT_VSS2 = '=16sIBBHI'


class VARIABLE_STORE_HEADER_VSS2(namedtuple('VARIABLE_STORE_HEADER_VSS2', 'Signature Size Format State Reserved Reserved1')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
EFI Variable Store
-----------------------------
Signature : {UUID(bytes_le=self.Signature)}
Size      : 0x{self.Size:08X} bytes
Format    : 0x{self.Format:02X}
State     : 0x{self.State:02X}
Reserved  : 0x{self.Reserved:04X}
Reserved1 : 0x{self.Reserved1:08X}
"""


VARIABLE_STORE_SIGNATURE_VSS2 = UUID('DDCF3617-3275-4164-98B6-FE85707FFE7D').bytes_le
VARIABLE_STORE_SIGNATURE_VSS2_AUTH = UUID('AAF32C78-947B-439A-A180-2E144EC37792').bytes_le

HDR_FMT_VSS = '<HBBIII16s'
# HDR_SIZE_VSS                  = struct.calcsize( HDR_FMT_VSS )
# NAME_OFFSET_IN_VAR_VSS        = HDR_SIZE_VSS


class EFI_HDR_VSS(namedtuple('EFI_HDR_VSS', 'StartId State Reserved Attributes NameSize DataSize guid')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
Header (VSS)
------------
VendorGuid : {{{EFI_GUID_STR(self.guid)}}}
StartId    : 0x{self.StartId:04X}
State      : 0x{self.State:02X}
Reserved   : 0x{self.Reserved:02X}
Attributes : 0x{self.Attributes:08X}
NameSize   : 0x{self.NameSize:08X}
DataSize   : 0x{self.DataSize:08X}
"""


HDR_FMT_VSS_AUTH = '<HBBIQQQIII16s'


class EFI_HDR_VSS_AUTH(namedtuple('EFI_HDR_VSS_AUTH', 'StartId State Reserved Attributes MonotonicCount TimeStamp1 TimeStamp2 PubKeyIndex NameSize DataSize guid')):
    __slots__ = ()
    # if you don't re-define __str__ method, initialize is to None
    # __str__ = None

    def __str__(self) -> str:
        return f"""
Header (VSS_AUTH)
----------------
VendorGuid     : {{{EFI_GUID_STR(self.guid)}}}
StartId        : 0x{self.StartId:04X}
State          : 0x{self.State:02X}
Reserved       : 0x{self.Reserved:02X}
Attributes     : 0x{self.Attributes:08X}
MonotonicCount : 0x{self.MonotonicCount:016X}
TimeStamp1     : 0x{self.TimeStamp1:016X}
TimeStamp2     : 0x{self.TimeStamp2:016X}
PubKeyIndex    : 0x{self.PubKeyIndex:08X}
NameSize       : 0x{self.NameSize:08X}
DataSize       : 0x{self.DataSize:08X}
"""


HDR_FMT_VSS_APPLE = '<HBBIII16sI'


class EFI_HDR_VSS_APPLE(namedtuple('EFI_HDR_VSS_APPLE', 'StartId State Reserved Attributes NameSize DataSize guid unknown')):
    __slots__ = ()

    def __str__(self) -> str:
        return f"""
Header (VSS_APPLE)
------------
VendorGuid : {{{EFI_GUID_STR(self.guid)}}}
StartId    : 0x{self.StartId:04X}
State      : 0x{self.State:02X}
Reserved   : 0x{self.Reserved:02X}
Attributes : 0x{self.Attributes:08X}
NameSize   : 0x{self.NameSize:08X}
DataSize   : 0x{self.DataSize:08X}
Unknown    : 0x{self.unknown:08X}
"""


def _getNVstore_VSS(nvram_buf: bytes, vss_type) -> Tuple[int, int, Union[VARIABLE_STORE_HEADER_VSS, VARIABLE_STORE_HEADER_VSS2, None]]:
    if vss_type == FWType.EFI_FW_TYPE_VSS2:
        sign = VARIABLE_STORE_SIGNATURE_VSS2
    elif vss_type == FWType.EFI_FW_TYPE_VSS2_AUTH:
        sign = VARIABLE_STORE_SIGNATURE_VSS2_AUTH
    else:
        sign = VARIABLE_STORE_SIGNATURE_VSS

    nvram_start = nvram_buf.find(sign)
    if -1 == nvram_start:
        return (-1, 0, None)
    buf = nvram_buf[nvram_start:]
    if (not isCorrectVSStype(buf, vss_type)):
        return (-1, 0, None)
    if vss_type in (FWType.EFI_FW_TYPE_VSS2, FWType.EFI_FW_TYPE_VSS2_AUTH):
        nvram_hdr = VARIABLE_STORE_HEADER_VSS2(*struct.unpack_from(VARIABLE_STORE_HEADER_FMT_VSS2, buf))
    else:
        nvram_hdr = VARIABLE_STORE_HEADER_VSS(*struct.unpack_from(VARIABLE_STORE_HEADER_FMT_VSS, buf))
    return (nvram_start, nvram_hdr.Size, nvram_hdr)


def getNVstore_VSS(nvram_buf: bytes):
    return _getNVstore_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS)


def getNVstore_VSS_AUTH(nvram_buf: bytes):
    return _getNVstore_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS_AUTH)


def getNVstore_VSS2(nvram_buf: bytes):
    return _getNVstore_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS2)


def getNVstore_VSS2_AUTH(nvram_buf: bytes):
    return _getNVstore_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS2_AUTH)


def getNVstore_VSS_APPLE(nvram_buf: bytes):
    return _getNVstore_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS_APPLE)


VSS_TYPES = (FWType.EFI_FW_TYPE_VSS, FWType.EFI_FW_TYPE_VSS_AUTH, FWType.EFI_FW_TYPE_VSS2, FWType.EFI_FW_TYPE_VSS2_AUTH, FWType.EFI_FW_TYPE_VSS_APPLE)
MAX_VSS_VAR_ALIGNMENT = 8


def isCorrectVSStype(nvram_buf: bytes, vss_type: str):
    if (vss_type not in VSS_TYPES):
        return False

    buf_size = len(nvram_buf)
    start = nvram_buf.find(VARIABLE_SIGNATURE_VSS)
    if (-1 == start):
        return False

    next_var = nvram_buf.find(VARIABLE_SIGNATURE_VSS, start + struct.calcsize(HDR_FMT_VSS))  # skip the minimum bytes required for the header
    if (-1 == next_var):
        next_var = buf_size

    buf_size -= start

    if (vss_type in (FWType.EFI_FW_TYPE_VSS, FWType.EFI_FW_TYPE_VSS2)):
        hdr_fmt = HDR_FMT_VSS
        efi_var_hdr = EFI_HDR_VSS(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))
    elif (vss_type in (FWType.EFI_FW_TYPE_VSS_AUTH, FWType.EFI_FW_TYPE_VSS2_AUTH)):
        hdr_fmt = HDR_FMT_VSS_AUTH
        efi_var_hdr = EFI_HDR_VSS_AUTH(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))
    elif (vss_type == FWType.EFI_FW_TYPE_VSS_APPLE):
        hdr_fmt = HDR_FMT_VSS_APPLE
        efi_var_hdr = EFI_HDR_VSS_APPLE(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))

    hdr_size = struct.calcsize(hdr_fmt)
    # check NameSize and DataSize
    name_offset = start + hdr_size
    if ((name_offset < next_var) and ((name_offset + efi_var_hdr.NameSize) < next_var)):
        valid_name = False
        if (efi_var_hdr.NameSize > 0):
            name = nvram_buf[name_offset: name_offset + efi_var_hdr.NameSize]
            try:
                name = name.decode("utf-16-le").split('\x00')[0]
                valid_name = defines.is_printable(name)
            except Exception as e:
                pass
        if (valid_name):
            end_var_offset = name_offset + efi_var_hdr.NameSize + efi_var_hdr.DataSize
            off_diff = next_var - end_var_offset
            if (off_diff == 0):
                return True
            elif (off_diff > 0):
                if (next_var == len(nvram_buf)) or (off_diff <= (MAX_VSS_VAR_ALIGNMENT - 1)):
                    return True
            else:
                if (next_var < len(nvram_buf)):
                    new_nex_var = nvram_buf.find(VARIABLE_SIGNATURE_VSS, next_var, next_var + len(VARIABLE_SIGNATURE_VSS) + (MAX_VSS_VAR_ALIGNMENT - 1))
                    if (new_nex_var != -1):
                        return True

    return False


def _getEFIvariables_VSS(nvram_buf: bytes, _fwtype: str) -> Dict[str, List[EfiVariableType]]:
    variables = dict()
    nvsize = len(nvram_buf)
    if _fwtype in (FWType.EFI_FW_TYPE_VSS, FWType.EFI_FW_TYPE_VSS2):
        hdr_fmt = HDR_FMT_VSS
    elif _fwtype in (FWType.EFI_FW_TYPE_VSS_AUTH, FWType.EFI_FW_TYPE_VSS2_AUTH):
        hdr_fmt = HDR_FMT_VSS_AUTH
    elif (FWType.EFI_FW_TYPE_VSS_APPLE == _fwtype):
        hdr_fmt = HDR_FMT_VSS_APPLE
    else:
        return variables
    hdr_size = struct.calcsize(hdr_fmt)
    start = nvram_buf.find(VARIABLE_SIGNATURE_VSS)
    if -1 == start:
        return variables

    while (start + hdr_size) < nvsize:
        efi_var_hdr = None
        variables = {}
        if _fwtype in (FWType.EFI_FW_TYPE_VSS, FWType.EFI_FW_TYPE_VSS2):
            efi_var_hdr = EFI_HDR_VSS(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))
        elif _fwtype in (FWType.EFI_FW_TYPE_VSS_AUTH, FWType.EFI_FW_TYPE_VSS2_AUTH):
            efi_var_hdr = EFI_HDR_VSS_AUTH(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))
        elif (FWType.EFI_FW_TYPE_VSS_APPLE == _fwtype):
            efi_var_hdr = EFI_HDR_VSS_APPLE(*struct.unpack_from(hdr_fmt, nvram_buf[start:]))

        if efi_var_hdr is None:
            return variables
        if (efi_var_hdr.StartId != VARIABLE_DATA):
            break

        if ((efi_var_hdr.State == 0xff) and (efi_var_hdr.DataSize == 0xffffffff) and (efi_var_hdr.NameSize == 0xffffffff) and (efi_var_hdr.Attributes == 0xffffffff)):
            name_size = 0
            data_size = 0
            # just skip variable with empty name and data for now
            next_var_offset = nvram_buf.find(VARIABLE_SIGNATURE_VSS, start + hdr_size, start + hdr_size + len(VARIABLE_SIGNATURE_VSS) + (MAX_VSS_VAR_ALIGNMENT - 1))
            if (next_var_offset == -1) or (next_var_offset > nvsize):
                break
        else:
            name_size = efi_var_hdr.NameSize
            data_size = efi_var_hdr.DataSize
            efi_var_name = "<not defined>"

            end_var_offset = start + hdr_size + name_size + data_size
            efi_var_buf = nvram_buf[start: end_var_offset]

            name_offset = hdr_size
            Name = efi_var_buf[name_offset: name_offset + name_size]
            if Name:
                efi_var_name = Name.decode("utf-16-le").split('\x00')[0]

            efi_var_data = efi_var_buf[name_offset + name_size: name_offset + name_size + data_size]
            guid = EFI_GUID_STR(efi_var_hdr.guid)
            if efi_var_name not in variables.keys():
                variables[efi_var_name] = []
            #                                off,   buf,         hdr,         data,         guid, attrs
            variables[efi_var_name].append((start, efi_var_buf, efi_var_hdr, efi_var_data, guid, efi_var_hdr.Attributes))

            # deal with different alignments (1-8)
            next_var_offset = nvram_buf.find(VARIABLE_SIGNATURE_VSS, end_var_offset, end_var_offset + len(VARIABLE_SIGNATURE_VSS) + (MAX_VSS_VAR_ALIGNMENT - 1))
            if (next_var_offset == -1) or (next_var_offset > nvsize):
                break

        if start >= next_var_offset:
            break
        start = next_var_offset

    return variables


def getEFIvariables_VSS(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS)


def getEFIvariables_VSS_AUTH(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS_AUTH)


def getEFIvariables_VSS2(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS2)


def getEFIvariables_VSS2_AUTH(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS2_AUTH)


def getEFIvariables_VSS_APPLE(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    return _getEFIvariables_VSS(nvram_buf, FWType.EFI_FW_TYPE_VSS_APPLE)


#######################################################################
#
# EVSA NVRAM (signature = 'EVSA')
#
#
VARIABLE_STORE_SIGNATURE_EVSA = b'EVSA'

TLV_HEADER = "<BBH"
tlv_h_size = struct.calcsize(TLV_HEADER)


def getNVstore_EVSA(nvram_buf: bytes) -> NvStore:
    l = (-1, -1, None)
    fv = NextFwVolume(nvram_buf)
    while fv is not None:
        if (fv.Guid == VARIABLE_STORE_FV_GUID):
            nvram_start = fv.Image.find(VARIABLE_STORE_SIGNATURE_EVSA)
            if (nvram_start != -1) and (nvram_start >= tlv_h_size):
                nvram_start = nvram_start - tlv_h_size
                l = (fv.Offset + nvram_start, fv.Size - nvram_start, None)
                break
        if (fv.Guid == ADDITIONAL_NV_STORE_GUID):
            nvram_start = fv.Image.find(VARIABLE_STORE_SIGNATURE_EVSA)
            if (nvram_start != -1) and (nvram_start >= tlv_h_size):
                nvram_start = nvram_start - tlv_h_size
                l = (fv.Offset + nvram_start, fv.Size - nvram_start, None)
        fv = NextFwVolume(nvram_buf, fv.Offset, fv.Size)
    return l


def EFIvar_EVSA(nvram_buf: bytes) -> Dict[str, List[EfiVariableType]]:
    image_size = len(nvram_buf)
    sn = 0
    EVSA_RECORD = "<IIII"
    evsa_rec_size = struct.calcsize(EVSA_RECORD)
    GUID_RECORD = "<H16s"
    guid_rc_size = struct.calcsize(GUID_RECORD)
    fof = 0
    variables = dict()
    while fof < image_size:
        fof = nvram_buf.find(VARIABLE_STORE_SIGNATURE_EVSA, fof)
        if fof == -1:
            break
        if fof < tlv_h_size:
            fof = fof + 4
            continue
        start = fof - tlv_h_size
        Tag0, Tag1, Size = struct.unpack(TLV_HEADER, nvram_buf[start: start + tlv_h_size])
        if Tag0 != 0xEC:  # Wrong EVSA block
            fof = fof + 4
            continue
        value = nvram_buf[start + tlv_h_size:start + Size]
        _, _, Length, _ = struct.unpack(EVSA_RECORD, value)
        if start + Length > image_size:  # Wrong EVSA record
            fof = fof + 4
            continue
        # NV storage EVSA found
        bof = 0
        guid_map = dict()
        var_list = list()
        value_list = dict()
        while (bof + tlv_h_size) < Length:
            Tag0, Tag1, Size = struct.unpack(TLV_HEADER, nvram_buf[start + bof: start + bof + tlv_h_size])
            if (Size < tlv_h_size):
                break
            value = nvram_buf[start + bof + tlv_h_size:start + bof + Size]
            bof = bof + Size
            if (Tag0 == 0xED) or (Tag0 == 0xE1):  # guid
                GuidId, guid0 = struct.unpack(GUID_RECORD, value)
                g = EFI_GUID_STR(guid0)
                guid_map[GuidId] = g
            elif (Tag0 == 0xEE) or (Tag0 == 0xE2):  # var name
                VAR_NAME_RECORD = f'<H{Size - tlv_h_size - 2:d}s'
                VarId, Name = struct.unpack(VAR_NAME_RECORD, value)
                Name = Name.decode("utf-16-le")[:-1]
                var_list.append((Name, VarId, Tag0, Tag1))
            elif (Tag0 == 0xEF) or (Tag0 == 0xE3) or (Tag0 == 0x83):  # values
                VAR_VALUE_RECORD = f'<HHI{Size - tlv_h_size - 8:d}s'
                GuidId, VarId, Attributes, Data = struct.unpack(VAR_VALUE_RECORD, value)
                value_list[VarId] = (GuidId, Attributes, Data, Tag0, Tag1)
            elif not ((Tag0 == 0xff) and (Tag1 == 0xff) and (Size == 0xffff)):
                pass
        var_count = len(var_list)
        var_list.sort()
        var1 = {}
        for i in var_list:
            name = i[0]
            VarId = i[1]
            # NameTag0 = i[2]
            # NameTag1 = i[3]
            if VarId in value_list:
                var_value = value_list[VarId]
            else:
                #  Value not found for VarId
                continue
            GuidId = var_value[0]
            guid = "NONE"
            if GuidId not in guid_map:
                # Guid not found for GuidId
                pass
            else:
                guid = guid_map[GuidId]
            if name not in variables.keys():
                variables[name] = []
            #                       off,   buf,  hdr,  data,         guid, attrs
            variables[name].append((start, b'', None, var_value[2], guid, var_value[1]))
        fof = fof + Length
    return variables


#
# Uncomment if you want to parse output buffer returned by NtEnumerateSystemEnvironmentValuesEx
# using 'chipsec_util uefi nvram' command
#
#
# Windows 8 NtEnumerateSystemEnvironmentValuesEx (infcls = 2)
#
# def guid_str(guid0, guid1, guid2, guid3):
#        return ( f'{guid0:08X}-{guid1:04X}-{guid2:04X}-{guid3[:2].encode('hex').upper():4}-{guid3[-6::].encode('hex').upper():6}')
#
# class EFI_HDR_WIN( namedtuple('EFI_HDR_WIN', 'Size DataOffset DataSize Attributes guid0 guid1 guid2 guid3') ):
#        __slots__ = ()
#        def __str__(self):
#            return f"""
# Header (Windows)
# ----------------
# VendorGuid= {{self.guid0:08X}-{self.guid1:04X}-{self.guid2:04X}-{self.guid3[:2].encode('hex').upper():4}-{self.guid3[-6::].encode('hex').upper():6}}
# Size      = 0x{self.Size:08X}
# DataOffset= 0x{self.DataOffset:08X}
# DataSize  = 0x{self.DataSize:08X}
# Attributes= 0x{self.Attributes:08X}
# """
"""
def getEFIvariables_NtEnumerateSystemEnvironmentValuesEx2( nvram_buf ):
        start = 0
        buffer = nvram_buf
        bsize = len(buffer)
        header_fmt = "<IIIIIHH8s"
        header_size = struct.calcsize( header_fmt )
        variables = dict()
        off = 0
        while (off + header_size) < bsize:
           efi_var_hdr = EFI_HDR_WIN( *struct.unpack_from( header_fmt, buffer[ off : off + header_size ] ) )

           next_var_offset = off + efi_var_hdr.Size
           efi_var_buf     = buffer[ off : next_var_offset ]
           efi_var_data    = buffer[ off + efi_var_hdr.DataOffset : off + efi_var_hdr.DataOffset + efi_var_hdr.DataSize ]

           #efi_var_name = "".join( buffer[ start + header_size : start + efi_var_hdr.DataOffset ] ).decode('utf-16-le')
           str_fmt = f'{efi_var_hdr.DataOffset - header_size:d}s'
           s, = struct.unpack( str_fmt, buffer[ off + header_size : off + efi_var_hdr.DataOffset ] )
           efi_var_name = unicode(s, "utf-16-le", errors="replace").split(u'\u0000')[0]

           if efi_var_name not in variables.keys():
               variables[efi_var_name] = []
           #                                off, buf,         hdr,         data,         guid,                                                                                 attrs
           variables[efi_var_name].append( (off, efi_var_buf, efi_var_hdr, efi_var_data, guid_str(efi_var_hdr.guid0, efi_var_hdr.guid1, efi_var_hdr.guid2, efi_var_hdr.guid3), efi_var_hdr.Attributes) )

           if 0 == efi_var_hdr.Size: break
           off = next_var_offset

        return variables
#    return ( start, next_var_offset, efi_var_buf, efi_var_hdr, efi_var_name, efi_var_data, guid_str(efi_var_hdr.guid0, efi_var_hdr.guid1, efi_var_hdr.guid2, efi_var_hdr.guid3), efi_var_hdr.Attributes )
"""


#
# Decoding S3 Resume Boot Script
#

class S3BootScriptType:
    EFI_BOOT_SCRIPT_TYPE_DEFAULT = 0x00
    EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT = 0xAA


def decode_s3bs_opcode(s3bootscript_type, script_data):
    if S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT == s3bootscript_type:
        return decode_s3bs_opcode_edkcompat(script_data)
    else:
        return decode_s3bs_opcode_def(script_data)


def encode_s3bs_opcode(s3bootscript_type: int, op: S3BOOTSCRIPT_ENTRY) -> bytes:
    if S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT == s3bootscript_type:
        return encode_s3bs_opcode_edkcompat(op)
    else:
        return encode_s3bs_opcode_def(op)


def decode_s3bs_opcode_def(data):
    opcode = None
    size = None
    width = None
    unknown = None
    count = None
    value = None
    mask = None

    op = None
    opcode, = struct.unpack('<B', data[: 1])
    try:
        logger().log_hal(script_opcodes[opcode])
    except:
        pass
    if S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_IO_WRITE_OPCODE == opcode:
        frmt = '<BBHIQ'
        size = struct.calcsize(frmt)
        opcode, width, address, alignment, count = struct.unpack(frmt, data[: size])
        op = op_io_pci_mem(opcode, size, width, address, unknown, count, data[size:], value, mask)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_IO_READ_WRITE_OPCODE == opcode:
        frmt = '<BBHIQQ'
        size = struct.calcsize(frmt)
        opcode, width, address, _, value, mask = struct.unpack(frmt, data[: size])
        op = op_io_pci_mem(opcode, size, width, address, unknown, count, None, value, mask)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_PCI_CONFIG_WRITE_OPCODE == opcode:
        frmt = '<BBHIQQ'
        size = struct.calcsize(frmt)
        opcode, width, unknown, _, address, count = struct.unpack(frmt, data[: size])
        op = op_io_pci_mem(opcode, size, width, address, unknown, count, data[size:], value, mask)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_PCI_CONFIG_READ_WRITE_OPCODE == opcode:
        frmt = '<BBHIQQQ'
        size = struct.calcsize(frmt)
        opcode, width, unknown, _, address, value, mask = struct.unpack(frmt, data[: size])
        op = op_io_pci_mem(opcode, size, width, address, unknown, count, None, value, mask)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_MEM_WRITE_OPCODE == opcode:
        frmt = '<BBHIQQ'
        size = struct.calcsize(frmt)
        opcode, width, unknown, _, address, count = struct.unpack(frmt, data[: size])
        op = op_io_pci_mem(opcode, size, width, address, unknown, count, data[size:], value, mask)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_MEM_READ_WRITE_OPCODE == opcode:
        frmt = '<BBHIQQQ'
        size = struct.calcsize(frmt)
        opcode, width, unknown, _, address, value, mask = struct.unpack(frmt, data[: size])
        op = op_io_pci_mem(opcode, size, width, address, unknown, count, None, value, mask)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_SMBUS_EXECUTE_OPCODE == opcode:
        frmt = '<BBQBB'
        size = struct.calcsize(frmt)
        opcode, address, command, operation, peccheck = struct.unpack(frmt, data[: size])
        op = op_smbus_execute(opcode, size, address, command, operation, peccheck)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_STALL_OPCODE == opcode:
        frmt = '<BBQ'
        size = struct.calcsize(frmt)
        opcode, _, duration = struct.unpack(frmt, data[: size])
        op = op_stall(opcode, size, duration)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_DISPATCH_OPCODE == opcode:
        frmt = '<BBHIQ'
        size = struct.calcsize(frmt)
        opcode, _, _, _, entrypoint = struct.unpack(frmt, data[: size])
        op = op_dispatch(opcode, size, entrypoint)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_DISPATCH_2_OPCODE == opcode:
        frmt = '<BBHIQQ'
        size = struct.calcsize(frmt)
        opcode, _, _, _, entrypoint, context = struct.unpack(frmt, data[: size])
        op = op_dispatch(opcode, size, entrypoint, context)
    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_TERMINATE_OPCODE == opcode:
        frmt = '<B'
        size = struct.calcsize(frmt)
        opcode, = struct.unpack(frmt, data[: size])
        op = op_terminate(opcode, size)
    else:
        op = op_unknown(opcode, 1)
        if logger().HAL:
            logger().log_warning(f'Unrecognized opcode {opcode:X}')

    return op

#
# @TODO: encode functions are not fully implemented
#


def encode_s3bs_opcode_def(op) -> bytes:
    encoded_opcode = b''

    if S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_IO_WRITE_OPCODE == op.opcode:
        encoded_hdr = struct.pack('<BBHIQ', op.opcode, op.width, op.address, 0x0, op.count)
        if op.values is None:
            encoded_opcode = encoded_hdr + op.buffer
        else:
            encoded_opcode = encoded_hdr + struct.pack(script_width_formats[op.width] * op.count, *op.values)

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_IO_READ_WRITE_OPCODE == op.opcode:
        encoded_opcode = struct.pack('<BBHIQQ', op.opcode, op.width, op.address, 0x0, op.value, op.mask)

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_PCI_CONFIG_WRITE_OPCODE == op.opcode or \
            S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_MEM_WRITE_OPCODE == op.opcode:
        encoded_hdr = struct.pack('<BBHIQQ', op.opcode, op.width, op.unknown, 0x0, op.address, op.count)
        if op.values is None:
            encoded_opcode = encoded_hdr + op.buffer
        else:
            encoded_opcode = encoded_hdr + struct.pack(script_width_formats[op.width] * op.count, *op.values)

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_PCI_CONFIG_READ_WRITE_OPCODE == op.opcode:
        frmt = '<BBHIQQQ'

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_MEM_READ_WRITE_OPCODE == op.opcode:
        encoded_opcode = struct.pack('<BBHIQQQ', op.opcode, op.width, op.unknown, 0x0, op.address, op.value, op.mask)

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_SMBUS_EXECUTE_OPCODE == op.opcode:
        frmt = '<BBQBB'

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_STALL_OPCODE == op.opcode:
        frmt = '<BBQ'

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_DISPATCH_OPCODE == op.opcode:
        encoded_opcode = struct.pack('<BBHIQ', op.opcode, 0x0, 0x0, 0x0, op.entrypoint)

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_DISPATCH_2_OPCODE == op.opcode:
        encoded_opcode = struct.pack('<BBHIQQ', op.opcode, 0x0, 0x0, 0x0, op.entrypoint, op.context)

    elif S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_TERMINATE_OPCODE == op.opcode:
        frmt = '<B'

    else:
        if logger().HAL:
            logger().log_warning(f'Unrecognized opcode {op.opcode:X}')

    return encoded_opcode


def decode_s3bs_opcode_edkcompat(data: bytes):
    opcode = None
    width = None
    count = None
    value = None
    mask = None

    op = None

    hdr_frmt = '<HB'
    header_size = struct.calcsize(hdr_frmt)
    opcode, size = struct.unpack(hdr_frmt, data[: header_size])
    opcode_data = data[header_size:]
    try:
        logger().log_hal(script_opcodes[opcode])
    except:
        pass

    if S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_IO_WRITE_OPCODE == opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_PCI_CONFIG_WRITE_OPCODE == opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_MEM_WRITE_OPCODE == opcode:

        frmt = '<IIQ'
        op_size = struct.calcsize(frmt)
        width, count, address = struct.unpack(frmt, opcode_data[: op_size])
        op = op_io_pci_mem(opcode, size, width, address, None, count, opcode_data[op_size:], value, mask)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_IO_READ_WRITE_OPCODE == opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_PCI_CONFIG_READ_WRITE_OPCODE == opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_MEM_READ_WRITE_OPCODE == opcode:
        frmt = '<IQ'
        sz = struct.calcsize(frmt)
        width, address = struct.unpack(frmt, opcode_data[: sz])
        frmt = 2 * script_width_formats[width]
        op_size = sz + struct.calcsize(frmt)
        value, mask = struct.unpack(frmt, opcode_data[sz: op_size])
        op = op_io_pci_mem(opcode, size, width, address, None, count, None, value, mask)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_SMBUS_EXECUTE_OPCODE == opcode:
        if logger().UTIL_TRACE or logger().HAL:
            logger().log_warning(f'Cannot parse opcode {opcode:X} yet')

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_STALL_OPCODE == opcode:
        frmt = '<Q'
        op_size = struct.calcsize(frmt)
        duration, = struct.unpack(frmt, opcode_data[: op_size])
        op = op_stall(opcode, size, duration)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_DISPATCH_OPCODE == opcode:
        frmt = '<Q'
        op_size = struct.calcsize(frmt)
        entrypoint, = struct.unpack(frmt, opcode_data[: op_size])
        op = op_dispatch(opcode, size, entrypoint)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_MEM_POLL_OPCODE == opcode:
        frmt = '<IQQQ'
        op_size = struct.calcsize(frmt)
        width, address, duration, looptimes = struct.unpack(frmt, opcode_data[: op_size])
        op = op_mem_poll(opcode, size, width, address, duration, looptimes)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_TERMINATE_OPCODE == opcode:
        op = op_terminate(opcode, size)

    else:
        op = op_unknown(opcode, size)
        if logger().HAL:
            logger().log_warning(f'Unrecognized opcode {opcode:X}')

    return op

#
# @TODO: encode functions are not fully implemented
#


def encode_s3bs_opcode_edkcompat(op: S3BOOTSCRIPT_ENTRY) -> bytes:
    encoded_opcode = b''

    if S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_IO_WRITE_OPCODE == op.opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_PCI_CONFIG_WRITE_OPCODE == op.opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_MEM_WRITE_OPCODE == op.opcode:

        encoded_hdr = struct.pack('<IIQ', op.width, op.count, op.address)
        if op.values is None:
            encoded_opcode = encoded_hdr + op.buffer
        else:
            encoded_opcode = encoded_hdr + struct.pack(script_width_formats[op.width] * op.count, *op.values)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_IO_READ_WRITE_OPCODE == op.opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_PCI_CONFIG_READ_WRITE_OPCODE == op.opcode or \
            S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_MEM_READ_WRITE_OPCODE == op.opcode:

        frmt = f'<IQ2{script_width_formats[op.width]}'
        encoded_opcode = struct.pack(frmt, op.width, op.address, op.value, op.mask)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_SMBUS_EXECUTE_OPCODE == op.opcode:
        pass

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_STALL_OPCODE == op.opcode:
        frmt = '<Q'

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_DISPATCH_OPCODE == op.opcode:
        encoded_opcode = struct.pack('<Q', op.entrypoint)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_MEM_POLL_OPCODE == op.opcode:
        encoded_opcode = struct.pack('<IQQQ', op.width, op.address, op.duration, op.looptimes)

    elif S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_TERMINATE_OPCODE == op.opcode:
        pass

    return encoded_opcode


def parse_s3bootscript_entry(s3bootscript_type: int, script: bytes, off: int, log_script: bool = False):
    entry_index = None
    entry_length = 0
    opcode = None
    entry_data = None

    remaining_len = len(script[off:])

    if S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT == s3bootscript_type:
        fhdr = '<HB'
        hdr_length = struct.calcsize(fhdr)
        if remaining_len < hdr_length:
            if logger().HAL:
                logger().log_warning(f'The script should have at least 0x{hdr_length:X} bytes to parse next entry')
            return (0, None)

        opcode, entry_length = struct.unpack(fhdr, script[off: off + hdr_length])
        if S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_TERMINATE_OPCODE == opcode:
            entry_length = hdr_length
        entry_data = script[off: off + entry_length]

        if entry_length > MAX_S3_BOOTSCRIPT_ENTRY_LENGTH:
            logger().log_error(f'[uefi] Unrecognized S3 boot script format (entry length = 0x{entry_length:X})')
            return (0, None)

        s3script_entry = S3BOOTSCRIPT_ENTRY(s3bootscript_type, entry_index, off, entry_length, entry_data)

    else:  # S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_DEFAULT

        fhdr = '<II'
        hdr_length = struct.calcsize(fhdr)
        f = fhdr + 'B'
        if remaining_len < (hdr_length + 1):
            if logger().HAL:
                logger().log_warning(f'The script should have at least 0x{hdr_length + 1:X} bytes to parse next entry')
            return (0, None)

        entry_index, entry_length, opcode = struct.unpack(f, script[off: off + hdr_length + 1])
        if S3BootScriptOpcode_MDE.EFI_BOOT_SCRIPT_TERMINATE_OPCODE == opcode:
            entry_length = hdr_length + 1
            entry_index = -1
        entry_data = script[off + hdr_length: off + entry_length]

        if entry_length > MAX_S3_BOOTSCRIPT_ENTRY_LENGTH:
            logger().log_error(f'[uefi] Unrecognized S3 boot script format (entry length = 0x{entry_length:X})')
            return (0, None)

        s3script_entry = S3BOOTSCRIPT_ENTRY(s3bootscript_type, entry_index, off, entry_length, entry_data)
        s3script_entry.header_length = hdr_length

    s3script_entry.decoded_opcode = decode_s3bs_opcode(s3bootscript_type, s3script_entry.data)

    if log_script:
        logger().log(str(s3script_entry))
    return (opcode, s3script_entry)


def encode_s3bootscript_entry(entry) -> Optional[bytes]:
    if S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT == entry.script_type:
        entry_hdr_buf = struct.pack('<HB', entry.decoded_opcode.opcode, entry.length)
    else:  # S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_DEFAULT
        entry_hdr_buf = struct.pack('<II', entry.index, entry.length)

    entry_val_buf = encode_s3bs_opcode(entry.script_type, entry.decoded_opcode)
    entry_buf = None
    if entry_val_buf is not None:
        entry_buf = entry_hdr_buf + entry_val_buf
    else:
        logger().log_warning(f'Could not encode opcode of boot script entry (type 0x{entry.script_type:X})')

    return entry_buf


def create_s3bootscript_entry_buffer(script_type: int, op, index=None) -> bytes:
    entry_val_buf = encode_s3bs_opcode(script_type, op)
    length = len(entry_val_buf)
    if S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT == script_type:
        length += struct.calcsize('<HB')
        entry_hdr_buf = struct.pack('<HB', op.opcode, length)
    else:  # S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_DEFAULT
        length += struct.calcsize('<II')
        entry_hdr_buf = struct.pack('<II', index, length)

    return (entry_hdr_buf + entry_val_buf)


def id_s3bootscript_type(script: bytes, log_script: bool = False) -> Tuple[int, int]:
    script_header_length = 0

    start_op, = struct.unpack('<B', script[: 1])
    if S3BootScriptOpcode_EdkCompat.EFI_BOOT_SCRIPT_TABLE_OPCODE == start_op:
        logger().log_hal('S3 Boot Script AA Parser')
        script_type = S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_EDKCOMPAT
        if log_script:
            logger().log(f'[uefi] Start opcode 0x{start_op:X}')
        # MdeModulePkg\Library\PiDxeS3BootScriptLib\BootScriptInternalFormat.h
        script_header_length = struct.calcsize("<HBHLHH")
    else:
        logger().log_hal('S3 Boot Script DEFAULT Parser')
        script_type = S3BootScriptType.EFI_BOOT_SCRIPT_TYPE_DEFAULT

    return (script_type, script_header_length)


#
# EFI Variable Header Dictionary
#
#
# Add your EFI variable details to the dictionary
#
# Fields:
# name          func_getefivariables            func_getnvstore
#
EFI_VAR_DICT: Dict[str, Dict[str, Any]] = {
    # UEFI
    FWType.EFI_FW_TYPE_UEFI: {'name': 'UEFI', 'func_getefivariables': getEFIvariables_UEFI, 'func_getnvstore': getNVstore_EFI},
    FWType.EFI_FW_TYPE_UEFI_AUTH: {'name': 'UEFI_AUTH', 'func_getefivariables': getEFIvariables_UEFI_AUTH, 'func_getnvstore': getNVstore_EFI_AUTH},
    # Windows 8 NtEnumerateSystemEnvironmentValuesEx (infcls = 2)
    # FWType.EFI_FW_TYPE_WIN     : {'name' : 'WIN',     'func_getefivariables' : getEFIvariables_NtEnumerateSystemEnvironmentValuesEx2, 'func_getnvstore' : None },
    # NVAR format
    FWType.EFI_FW_TYPE_NVAR: {'name': 'NVAR', 'func_getefivariables': getEFIvariables_NVAR, 'func_getnvstore': getNVstore_NVAR},
    # $VSS NVRAM format
    FWType.EFI_FW_TYPE_VSS: {'name': 'VSS', 'func_getefivariables': getEFIvariables_VSS, 'func_getnvstore': getNVstore_VSS},
    # $VSS Authenticated NVRAM format
    FWType.EFI_FW_TYPE_VSS_AUTH: {'name': 'VSS_AUTH', 'func_getefivariables': getEFIvariables_VSS_AUTH, 'func_getnvstore': getNVstore_VSS_AUTH},
    # VSS2 NVRAM format
    FWType.EFI_FW_TYPE_VSS2: {'name': 'VSS2', 'func_getefivariables': getEFIvariables_VSS2, 'func_getnvstore': getNVstore_VSS2},
    # VSS2 Authenticated NVRAM format
    FWType.EFI_FW_TYPE_VSS2_AUTH: {'name': 'VSS2_AUTH', 'func_getefivariables': getEFIvariables_VSS2_AUTH, 'func_getnvstore': getNVstore_VSS2_AUTH},
    # Apple $VSS formart
    FWType.EFI_FW_TYPE_VSS_APPLE: {'name': 'VSS_APPLE', 'func_getefivariables': getEFIvariables_VSS_APPLE, 'func_getnvstore': getNVstore_VSS_APPLE},
    # EVSA
    FWType.EFI_FW_TYPE_EVSA: {'name': 'EVSA', 'func_getefivariables': EFIvar_EVSA, 'func_getnvstore': getNVstore_EVSA},
}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

# -------------------------------------------------------------------------------
#
# CHIPSEC: Platform Hardware Security Assessment Framework
#
# -------------------------------------------------------------------------------

"""
CMOS memory specific functions (dump, read/write)

usage:
    >>> cmos.dump_low()
    >>> cmos.dump_high()
    >>> cmos.dump()
    >>> cmos.read_cmos_low( offset )
    >>> cmos.write_cmos_low( offset, value )
    >>> cmos.read_cmos_high( offset )
    >>> cmos.write_cmos_high( offset, value )
"""
from typing import List
from chipsec.hal import hal_base
import chipsec.library.logger

CMOS_ADDR_PORT_LOW = 0x70
CMOS_DATA_PORT_LOW = 0x71
CMOS_ADDR_PORT_HIGH = 0x72
CMOS_DATA_PORT_HIGH = 0x73


class CMOS(hal_base.HALBase):

    def __init__(self, cs):
        super(CMOS, self).__init__(cs)

    def read_cmos_high(self, offset: int) -> int:
        self.cs.io.write_port_byte(CMOS_ADDR_PORT_HIGH, offset)
        return self.cs.io.read_port_byte(CMOS_DATA_PORT_HIGH)

    def write_cmos_high(self, offset: int, value: int) -> None:
        self.cs.io.write_port_byte(CMOS_ADDR_PORT_HIGH, offset)
        self.cs.io.write_port_byte(CMOS_DATA_PORT_HIGH, value)

    def read_cmos_low(self, offset: int) -> int:
        self.cs.io.write_port_byte(CMOS_ADDR_PORT_LOW, 0x80 | offset)
        return self.cs.io.read_port_byte(CMOS_DATA_PORT_LOW)

    def write_cmos_low(self, offset: int, value: int) -> None:
        self.cs.io.write_port_byte(CMOS_ADDR_PORT_LOW, offset)
        self.cs.io.write_port_byte(CMOS_DATA_PORT_LOW, value)

    def dump_low(self) -> List[int]:
        cmos_buf = [0xFF] * 0x80
        orig = self.cs.io.read_port_byte(CMOS_ADDR_PORT_LOW)
        for off in range(0x80):
            cmos_buf[off] = self.read_cmos_low(off)
        self.cs.io.write_port_byte(CMOS_ADDR_PORT_LOW, orig)
        return cmos_buf

    def dump_high(self) -> List[int]:
        cmos_buf = [0xFF] * 0x80
        orig = self.cs.io.read_port_byte(CMOS_ADDR_PORT_HIGH)
        for off in range(0x80):
            cmos_buf[off] = self.read_cmos_high(off)
        self.cs.io.write_port_byte(CMOS_ADDR_PORT_HIGH, orig)
        return cmos_buf

    def dump(self) -> None:
        self.logger.log("Low CMOS memory contents:")
        chipsec.library.logger.pretty_print_hex_buffer(self.dump_low())
        self.logger.log("\nHigh CMOS memory contents:")
        chipsec.library.logger.pretty_print_hex_buffer(self.dump_high())
